#!/usr/bin/env python
from __future__ import print_function
from builtins import map
from builtins import str
from builtins import filter
from builtins import object
import sys

"""Code for handling RNA secondary structure.

RNA secondary structures can be represented in many different ways. The
representation makes a large difference to the efficiency of different
algorithms, so several different structural representations (and the means
to interconvert them) are provided.

Provides following classes:
    Stem: representation of a stem in a secondary structure
    Partners: list holding partner of each position.
    PseudoknowRemover: detects and removes pseudoknots.
    Pairs: list of base pairs in a structure.
    StructureString: string holding a secondary structure representation.
    ViennaStructure: representation of Vienna-format RNA structure.
    WussStructure: Wash U secondary structure format, handles pseudoknots.
    StructureNode: for tree representation of nested RNA structure.
"""


class keep_chars(object):
    """Returns a filter object o(s): call to return a filtered string.

    Specifically, strips out everything in s that is not in keep.
    This filter is case sensitive by default.
    """
    allchars = str.maketrans('', '')

    def __init__(self, keep, case_sens=True):
        """Returns a new keep_chars object, based on string keep"""
        if not case_sens:
            low = keep.lower()
            up = keep.upper()
            keep = low + up
        self.delchars = ''.join([c for c in self.allchars if c not in keep])

    def __call__(self, s, a=None, d=None):
        """f(s) -> s, translates using self.allchars and self.delchars"""
        if a is None:
            a = self.allchars
        if d is None:
            d = self.delchars
        return s.translate(a, d)


maybe_number = keep_chars('0123456789.+-eE')


def float_from_string(data):
    """Extracts a floating point number from string in data, if possible."""
    return float(maybe_number(data))


class PairError(ValueError):
    """Base class for errors in pairing."""
    pass


class Partners(list):
    """Holds list p such that p[i] is the index of the partner of i, or None.

    Primarily useful for testing whether a specified base is paired and, if so,
    extracting its partner.

    Each base may have precisely 0 or 1 partners. If A pairs with B, B must
    pair with A.
    All inconsistencies will be removed by setting previous partners to None.
    Checking for conflicts and raising errors should be done in method that
    constructs the Partners.

    If constructing by hand, should initialize with list of [None] * seq_length.
    Typically, Partners will be constructed by code from some other data. Use
    the EmptyPartners(n) factory function to get an empty Partners list of
    length n.
    """

    def __setitem__(self, index, item):
        """Sets self[index] to item, enforcing integrity constraints."""
        if index == item:
            raise ValueError("Cannot set base %s to pair with itself." % item)
        # if item already paired, raise Error or make partner unpaired
        if item and self[item]:
            self[self[item]] = None
        # if already paired, need to make partner unpaired
        curr_partner = self[index]
        if curr_partner is not None:
            list.__setitem__(self, curr_partner, None)
        # set self[index] to item
        list.__setitem__(self, index, item)
        # if item is not None, set self[item] to index
        if item is not None:
            list.__setitem__(self, item, index)

    def toPairs(self):
        """Converts the partners to sorted list of pairs."""
        result = Pairs()
        for first, second in enumerate(self):
            if first < second:
                result.append((first, second))
        return result

    def _not_implemented(self, *args, **kwargs):
        """Raises NotImplementedError for 'naughty' methods.

        Not allowed any methods that insert/remove items or that change the
        order of the items, including things like sort or reverse.
        """
        raise NotImplementedError

    __delitem__ = __delslice__ = __iadd__ = __imul__ = __setslice__ = append \
        = extend = insert = pop = remove = reverse = sort = _not_implemented


def EmptyPartners(length):
    """Returns empty list of Partners with specified length."""
    return Partners([None] * length)


class Pairs(list):
    """Holds list of base pairs, each of which is a 2-element sequence.

    This is a very lightweight object for storing base pairs, and does not
    perform any validation. Useful as an intermediate in many different
    calculations.
    """

    def toPartners(self, length, offset=0, strict=True):
        """Returns a Partners object, if possible.

        length of resulting sequence must be specified.
        offset is optional, and is added to each index.
        strict specifies whether collisions cause fatal errors. if not strict
        conflicts will be removed by the Partners object.
        """
        result = EmptyPartners(length)
        for up, down in self:
            upstream = up + offset
            downstream = down + offset

            if result[upstream] or result[downstream]:
                if strict:
                    raise ValueError("Pairs contain conflicting partners: %s"
                                     % self)
            result[upstream] = downstream
        return result

    def toVienna(self, length, offset=0, strict=True):
        """Returns a Vienna structure string, if possible.

        length of resulting sequence must be specified.
            Instead of parsing the sequence length, you can also throw in
            an object that has the required length (such as the sequence that
            the structure corresponds to).
        offset is optional, and is added to each index.
        strict specifies whether collisions cause fatal errors.
        """
        if self.hasPseudoknots():
            raise Exception("Pairs contains pseudoknots %s" % (self))

        try:
            length = int(length)
        except ValueError:  # raised when length can't be converted to int
            length = len(length)

        p = self.directed()
        result = ['.'] * length
        for up, down in p:
            try:
                upstream = up + offset
                downstream = down + offset
            except TypeError:
                continue

            if strict:
                if (result[upstream] != '.') or (result[downstream] != '.'):
                    raise ValueError("Pairs contain conflicting partners: %s"
                                     % self)
            result[upstream] = '('
            result[downstream] = ')'
        return ViennaStructure(''.join(result))

    def tuples(self):
        """Converts all pairs in self to tuples, in place.

        Useful for constructing dicts and for sorting (otherwise, pairs of
        different types, e.g. lists and tuples, will sort according to type
        rather than to position).
        """
        self[:] = list(map(tuple, self))

    def unique(self):
        """Returns copy of self omitting duplicate pairs, preserving order.

        Keeps the first occurrence of each pair.
        """
        seen = {}
        result = []
        for p in map(tuple, self):
            if p not in seen:
                seen[p] = True
                result.append(p)
        return Pairs(result)

    def directed(self):
        """Returns copy of self where all pairs are (upstream, downstream).

        Omits any unpaired bases and any duplicates. Result is in arbitrary
        order.
        """
        seen = {}
        for up, down in self:
            if (up is None) or (down is None):
                continue  # omit unpaired bases
            if up > down:
                up, down = down, up
            seen[(up, down)] = True
        result = list(seen.keys())
        return Pairs(result)

    def symmetric(self):
        """Retruns copy of self where  each up, down pair has a down, up pair.

        Result is in arbitrary order. Double pairs and pairs containing None
        are left out.
        """
        result = self.directed()
        result.extend([(down, up) for up, down in result])
        return Pairs(result)

    def paired(self):
        """Returns copy of self omitting items where a 'partner' is None."""
        return Pairs(list(filter(not_none, self)))

    def hasPseudoknots(self):
        """Returns True if the pair list contains pseudoknots.

        (good_up,good_down) <=> (checked_up,checked_down)
        pseudoknot if checked_up<good_down and checked_down>good_down
        """
        pairs = self.directed()
        seen = []  # list of pairs against which you compare each time
        pairs.sort()
        for pair in pairs:
            if not seen:
                seen.append(pair)
            else:
                lastseen_up, lastseen_down = seen[-1]
                while pair[0] > lastseen_down:
                    seen.pop()
                    if not seen:
                        break
                    else:
                        lastseen_up, lastseen_down = seen[-1]
                if not seen:
                    seen.append(pair)
                    continue
                if pair[1] > lastseen_down:
                    # pseudoknot found
                    return True
                else:
                    # good pair
                    seen.append(pair)
        return False

    def hasConflicts(self):
        """Returns True if the pair list contains conflicts.

        Conflict occurs when a base has two different partners, or is asserted
        to be both paired and unpaired.
        """
        partners = {}
        for first, second in self:
            #print >>sys.stderr, "first:", first, "second:", second
            if first is None:
                if second is None:
                    continue  # no pairing info
                else:
                    first, second = second, first  # swap order so None is 2nd
            if second is None:  # check first isn't paired
                if partners.get(first, None) is not None:
                    return True
                else:
                    partners[first] = None
            else:  # first and second were both non-empty: check partners
                if first in partners:
                    if partners[first] != second:
                        print("here2", file=sys.stderr)
                        print("first:", first, "second:", second,
                              "partners[first]", partners[first], file=sys.stderr)
                        print("partners:", partners)
                        return True
                if second in partners:
                    if partners[second] != first:
                        print("here3", file=sys.stderr)
                        print("first:", first, "second:", second,
                              "partners[second]:", partners[second], file=sys.stderr)
                        return True
                # add current pair to the list of constraints
                partners[first] = second
                partners[second] = first
        # can only get here if there weren't conflicts
        return False

    def mismatches(self, sequence, pairs=None):
        """Counts pairs that can't form in sequence.

        Sequence must have a Pairs property that acts like a dictionary
        containing a 2-element tuple for each valid pair. Can also pass in
        the pairs explicitly.
        """
        mismatches = 0
        if pairs is None:
            try:
                pairs = sequence.Alphabet.Pairs
            except AttributeError:
                pairs = sequence.Pairs

        for up, down in self.directed():
            curr = (sequence[up], sequence[down])
            if curr not in pairs:
                mismatches += 1
        return mismatches


class StructureString(str):
    """Base class for ViennaStructure and WussStructure. Immutable.

    StructureString holds a structure and a energy. By default energy is
    set to None.
    If you compare two StructureStrings the structure is the only important
    thing, since the energy is relative to the associated sequence.
    """
    Alphabet = None
    StartSymbols = ''  # dict of symbols that start base pairs
    EndSymbols = ''  # dict of symbols that end base pairs

    def __new__(cls, Structure, Energy=None):
        """Returns new StructureString."""
        a = cls.Alphabet
        if a:
            for i in Structure:
                if i not in a:
                    raise ValueError(
                        "Tried to include unknown symbol '%s'" % i)

        return str.__new__(cls, Structure)

    def __init__(self, Structure='', Energy=None):
        """Initializes StructureString with Structure and optionally Energy."""
        self.Energy = Energy
        self.toPartners()

    def __str__(self):
        """Returns string representaion of structure and energy, if known.

        Energy = 0 is different from Energy = None. Latter case is not printed.
        """
        if not self.Energy == None:
            return self + ' (' + str(self.Energy) + ')'
        else:
            return str.__str__(self)

    def toPartners(self):
        """Makes list containing partner of each position.

        Constructs a list from 0 to the number of bases, where each position
        contains the index of its pair (or None if it is unpaired).

        Note that the numbering starts at 0 for the first position.

        The algorithm here relies on the fact that any completely nested
        base-paired structure (no pseudoknots!) can be formally represented
        as a tree. Consequently, when you hit a closed base pair, you know
        that it it must pair with the last base pair you opened.
        """
        num_bases = len(self)  # number of bases
        result = [None] * len(self)  # array of None, one for each base
        stack = []
        start = self.StartSymbols
        end = self.EndSymbols
        for i, symbol in enumerate(self):
            if symbol in start:  # open a pair
                stack.append(i)
            elif symbol in end:  # close a pair
                curr = stack.pop()  # return and delete last element
                result[i] = curr  # make i pair with the last element...
                result[curr] = i  # ...and the last element pair with i

        # test whether there are any open pairs left unaccounted for
        if stack:
            raise IndexError("Too many open pairs in structure:\n%s" % self)
        return Partners(result)

    def toPairs(self):
        """Makes list of (upstream,downstream) partners.

        Note that the numbering starts at 0 for the first position.
        Key will always be smaller than value.

        Result is in arbitrary order.
        """
        result = {}
        stack = []
        start = self.StartSymbols
        end = self.EndSymbols
        for i, symbol in enumerate(self):
            if symbol in start:  # open a pair
                stack.append(i)
            elif symbol in end:  # close a pair
                result[stack.pop()] = i
        # test whether there are any open pairs left unaccounted for
        if stack:
            raise IndexError("Too many open pairs in structure:\n%s" % self)
        return Pairs([(key, result[key]) for key in result])


class ViennaStructure(StructureString):
    """Contains a Vienna dot-bracket structure, possibly with energy."""
    Alphabet = dict.fromkeys('(.)')
    StartSymbols = {'(': None}  # dict of symbols that start base pairs
    EndSymbols = {')': None}  # dict of symbols that end base pairs


def Vienna(data, Energy=None):
    """Tries to extract structure and energy from string data.

    Returns (structure, energy) where energy might be None.

    structure is just anything before the first space: doesn't validate.
    """
    pieces = data.strip().split(None, 1)
    if not pieces:
        return ViennaStructure('', Energy)
    else:
        if not Energy:
            try:
                energy = float_from_string(pieces[1])
            except (TypeError, ValueError, IndexError):
                energy = Energy
        else:  # energy given by user overrules the one in structure
            energy = Energy
    return ViennaStructure(pieces[0], energy)



#!/usr/bin/python

"""
This code is still experimental
"""

from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import (ascii, bytes, chr, dict, filter, hex, input,
                      int, map, next, oct, open, pow, range, round,
                      str, super, zip, object)  # future package

from collections import defaultdict, Sequence
try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping
import itertools as it
import numpy as np
from sklearn.cluster import DBSCAN
from sklearn.manifold import MDS
from sklearn.decomposition import PCA
import time
import forgi.threedee.model.descriptors as ftmd
import forgi.threedee.utilities.vector as ftuv
import scipy.stats
import matplotlib.pyplot as plt
import warnings
from scipy.sparse import lil_matrix
import forgi.threedee.model.similarity as ftms
import pandas as pd
import logging
log = logging.getLogger(__name__)

np_nans = lambda *args, **kwargs: np.ones(*args, **kwargs) * np.nan


class RMSDMatrix(Sequence):
    def __init__(self, shape):
        self._matrix = lil_matrix(shape, dtype=float)
        self._len = shape[0]

    def __setitem__(self, key, value):
        try:
            log.info("Setting {} to array with shape {}".format(key, value.shape))
        except:
            pass
        self._matrix[key] = value + 1

    def __getitem__(self, key):
        return self._matrix[key] - 1

    def __len__(self):
        return self._len


class Ensemble(Mapping):
    # INITIALIZATION
    def __init__(self, cgs, reference_cg=None, sort_key=None):
        """
        An Ensemble is a sequence of Coarse grained RNAs, all of which must correspond
                    to the same RNA 2D structure.

        :param cgs: An iterable of coarse grain RNAs or a mapping key->cg,
                    from which the ensemble is constructed.
                    The ensemble may keep references to some of the cgs and
                    modify them by centering them on the centroid.
                    The ensemble is not guaranteed to keep/copy/reference any of the
                    cg's properties except the coords and twists.
        :param sort_key: Optional. A function that takes a cg (if cgs is a sequence)
                         or a key (if cgs is a mapping)
        """
        self._reference_cg = reference_cg

        self._cgs = []
        self._cg_lookup = {}
        # The order of cgs, as specified by sort_key
        self._cg_sequence = []
        self._cg_rev_lookup = defaultdict(list)
        if isinstance(cgs, Mapping):
            for key in sorted(cgs.keys(), key=sort_key):
                self._add_to_cg_list(cgs[key], key)
        else:
            if sort_key is not None:
                ids = list(
                    sorted(range(len(cgs)), key=lambda x: sort_key(cgs[x])))
            else:
                ids = list(range(len(cgs)))

            for key in ids:
                self._add_to_cg_list(cgs[key], key)
        # Center all cg coords on their centroid
        for cg in self._cgs:
            cg.coords.center()
        if self._reference_cg is not None:
            self._reference_cg.coords.center()

        ############## Caching of some descriptors ############################

        # The rmsd matrix. nan means the value needs to be calculated and will then be stored here.
        self._rmsd = RMSDMatrix((len(self._cgs), len(self._cgs)))
        for i in range(len(self._rmsd)):
            self._rmsd[i, i] = 0.0
        # 1D descriptors
        self._descriptors = {}

    def _get_descriptor(self, descr_name):
        if descr_name not in self._descriptors:
            self._descriptors[descr_name] = calculate_descriptor_for(descr_name,
                                                                     self._cgs,
                                                                     *self._get_args_for(descr_name))
            # if descr_name == "rmsd_to_last":
            #    self._rmsd[-1,:]=self._descriptors[descr_name]
            #    self._rmsd[:, -1]=self._descriptors[descr_name]

        return self._descriptors[descr_name]

    def _add_to_cg_list(self, cg, key):
        """
        During construction of the ensemble, this is used.

        Save some bookkeeping variables and store the cg,
        if it is not identical to the previouse.

        :param cg: The coarse grained RNA
        :param key: The index or name under which this cg can be retrieved again.
        """
        log.debug("Adding cg {} with key {}".format(cg.name, key))
        # In the future, we might have to check the rmsd in addition,
        # if the MCMC will allow rotation/translation
        if not self._cgs or cg.coords != self._cgs[-1].coords or cg.twists != self._cgs[-1].twists:
            if self._cgs:
                log.debug("Coords or twists different!")
            else:
                log.debug("First cg added")
            self._cgs.append(cg)
        else:
            log.debug("It is the same as the previous")
        # Else: only store lookup entry pointing to previous cg
        self._cg_lookup[key] = len(self._cgs) - 1
        self._cg_rev_lookup[len(self._cgs) - 1].append(key)
        self._cg_sequence.append(len(self._cgs) - 1)

    # MEMBER LOOKUP
    def __getitem__(self, key):
        """Using a key like used upon ensemble construction, retrieve the corresponding element"""
        return self._cgs[self._cg_lookup[key]]

    def __iter__(self):
        return iter(self._cg_lookup.keys())

    def __len__(self):
        return len(self._cgs)

    def at_timestep(self, timestep):
        """
        Access an ensemble member by the timestep.

        In contrast to __getitem__, which uses the index supplied upon ensemble creation,
        this uses the sort order specified by sort_key in the __init__ function.
        It is useful if the ensemble is seen as a trajectory.
        This method gives the ith frame of this trajectory.

        :param timestep: The number of the frame (cg) in the trajectory that should be retrieved.
        :returns: A coarse-grained RNA.
        """
        if hasattr(timestep, "__len__") and len(timestep) <= 3:
            seq = []
            for i in range(*timestep):
                seq.append(self._cgs[self._cg_sequence[i]])
            return seq
        return self._cgs[self._cg_sequence[timestep]]
    # RMSD and RMSD based calculations

    def rmsd_between(self, key1, key2, mode="key"):
        """
        Return (and cache) the rmsd between two structures.

        :param key1, key2: Two keys to reference two cgs
        :param mode: "key" or "timestep". Whether the keys are timesteps or the keys
                     used by __getitem__
        :returns: the rmsd as float
        """
        if mode == "key":
            i = self._cg_lookup[key1]
            j = self._cg_lookup[key2]
        elif mode == "timestep":
            i = self._cg_sequence[key1]
            j = self._cg_sequence[key2]
        else:
            raise ValueError("Invalid mode {}".format(mode))
        if self._rmsd[i, j] < 0:
            self._rmsd[i, j] = self._rmsd[j, i] = self._cgs[i].coords.rmsd_to(
                self._cgs[j].coords)
        return self._rmsd[i, j]

    def _calculate_complete_rmsd_matrix(self):
        """
        Fill out all empty fields in the rmsd matrix.
        """
        if np.any(self._rmsd < 0):
            # print(np.where(np.isnan(self._rmsd)))
            log.info("Starting complete rmsd calculation at {}".format(time.time()))
            for i, j in it.combinations(range(len(self)), 2):
                if self._rmsd[i, j] < 0:
                    self._rmsd[i, j] = self._rmsd[j, i] = self._cgs[i].coords.rmsd_to(
                        self._cgs[j].coords)
            log.info("Finished complete rmsd calculation at {}".format(time.time()))

    def _cluster_dbscan(self):
        """"
        Cluster all structures based on the DBSCAN algorithm
        using the pairwise RMSD as distance.
        """
        self._calculate_complete_rmsd_matrix()
        db = DBSCAN(eps=np.mean(
            self._rmsd[0]) / 3, min_samples=2, metric="precomputed").fit(self._rmsd)
        return db

    def _get_args_for(self, descriptor_name):
        """
        Get the arguments that are required to calculate the given descriptor of the ensemble's cgs.

        :param descriptor_name: The name of the descriptor to be calculated.
        """
        if descriptor_name == "rmsd_to_reference":
            if self._reference_cg:
                return [self._reference_cg]
            else:
                return [self._cgs[0]]
        elif descriptor_name == "rmsd_to_last":
            return [self._cgs[-1]]
        else:
            return []

    def view_delta_rmsd_vs_steps(self):
        self._calculate_complete_rmsd_matrix()
        fig, axes = plt.subplots(2)
        a_rmsd = np_nans(len(self._cg_sequence) // 2)
        min_rmsd = np_nans(len(self._cg_sequence) // 2)
        max_rmsd = np_nans(len(self._cg_sequence) // 2)
        for d in range(len(a_rmsd)):
            l = [self._rmsd[self._cg_sequence[i], self._cg_sequence[i + d]]
                 for i in range(len(self._cg_sequence) - d)]
            a_rmsd[d] = sum(l) / len(l)
            min_rmsd[d] = min(l)
            max_rmsd[d] = max(l)
        for ax in axes:
            ax.set_xlabel("Steps apart")
            ax.set_ylabel("Average RMSD")
            ax.plot(list(range(len(a_rmsd))), a_rmsd, label="Average RMSD")
            ax.plot(list(range(len(min_rmsd))), min_rmsd, label="Minimal RMSD")
            ax.plot(list(range(len(max_rmsd))), max_rmsd, label="Maximal RMSD")
            ax.plot([0, len(max_rmsd)], [np.max(self._rmsd), np.max(
                self._rmsd)], "-.", label="Maximal RMSD in whole simulation")
            ax.plot([0, len(max_rmsd)], [np.mean(self._rmsd), np.mean(
                self._rmsd)], "-.", label="Average RMSD in whole simulation")
            ax.legend(prop={'size': 6})
        axes[1].set_xlim([0, 50])

        plt.savefig("rmsd_steps_apart_{}.svg".format(self._cgs[0].name))

        plt.clf()
        plt.close()

    def view_2d_embedding(self, reference=None):
        # http://baoilleach.blogspot.co.at/2014/01/convert-distance-matrix-to-2d.html

        if reference is None:
            # First cluster all structures based on pairwise RMSD
            db = self._cluster_dbscan()
            labels = db.labels_
            core_samples_mask = np.zeros_like(db.labels_, dtype=bool)
            core_samples_mask[db.core_sample_indices_] = True
            unique_labels = set(labels)

            # Then calculate the 2D coordinates for our embedding
            mds = MDS(n_components=2,
                      dissimilarity="precomputed", random_state=6)
            results = mds.fit(self._rmsd)
            coords = results.embedding_

            # Now plot
            plt.plot(coords[:, 0], coords[:, 1], '-', color="blue")

            colors = plt.cm.Spectral(np.linspace(0, 1, len(unique_labels)))
            for k, col in zip(unique_labels, colors):
                if k == -1:
                    # Black used for noise.
                    col = 'k'
                class_member_mask = (labels == k)

                plt.plot(coords[:, 0][class_member_mask & core_samples_mask],
                         coords[:, 1][class_member_mask & core_samples_mask],
                         'o', markerfacecolor=col, markeredgecolor='k', markersize=6
                         )
                plt.plot(coords[:, 0][class_member_mask & ~core_samples_mask],
                         coords[:, 1][class_member_mask & ~core_samples_mask],
                         'o', markerfacecolor=col, markeredgecolor=col, markersize=1
                         )
            plt.savefig("embedding_{}.svg".format(self._cgs[0].name))
            plt.clf()
            plt.close()
        else:
            # Create a huge distance matrix
            alldists = np.zeros(
                ((len(self._cgs) + len(reference) + 1), (len(self._cgs) + len(reference) + 1)))
            for i, j in it.combinations(range(len(alldists)), 2):
                if i < len(self._cgs):
                    cg1 = self._cgs[i]
                elif i < len(self._cgs) + len(reference):
                    cg1 = reference[i - len(self._cgs)]
                else:
                    assert i == len(self._cgs) + len(reference)
                    cg1 = self._reference_cg
                if j < len(self._cgs):
                    cg2 = self._cgs[j]
                elif j < len(self._cgs) + len(reference):
                    cg2 = reference[j - len(self._cgs)]
                else:
                    assert j == len(self._cgs) + len(reference)
                    cg2 = self._reference_cg
                alldists[i, j] = alldists[j, i] = ftms.cg_rmsd(cg1, cg2)
            # Then calculate the 2D coordinates for our embedding
            mds = MDS(n_components=2,
                      dissimilarity="precomputed", random_state=6)
            results = mds.fit(alldists)
            coords = results.embedding_
            # Now plot
            plt.plot(coords[len(self._cgs):len(self._cgs) + len(reference), 0],
                     coords[len(self._cgs):len(self._cgs) + len(reference), 1], 's', color="green")
            plt.plot(coords[:len(self._cgs), 0],
                     coords[:len(self._cgs), 1], '-o', color="blue")
            plt.plot([coords[-1, 0]], [coords[-1, 1]], 's', color="red")
            plt.savefig("embedding1_{}.svg".format(self._cgs[0].name))
            plt.clf()
            plt.close()

    def color_by_energy(self, bins, ref_ensemble=None, ref_energies=None,
                        x="rmsd_to_reference", y="rmsd_to_last"):
        """"""
        plt.set_cmap('jet_r')

        # Get the data for plotting of the ensemble
        data_x = self._get_descriptor(x)
        data_y = self._get_descriptor(y)
        energies = self._get_descriptor("info_energy")
        image = get_energy_image(data_x, data_y, energies, bins)
        # Label the plot
        plt.xlabel(x)
        plt.ylabel(y)
        plt.imshow(image.T, interpolation="nearest", origin='lower',
                   extent=[bins[0][0], bins[0][-1], bins[1][0], bins[1][-1]],
                   aspect='auto')
        plt.xlim([bins[0][0], bins[0][-1]])
        plt.ylim([bins[1][0], bins[1][-1]])
        plt.colorbar()
        figname = "minEnergy_{}_{}_{}.svg".format(self._cgs[0].name, x, y)
        plt.savefig(figname)
        log.info("Figure {} created".format(figname))
        plt.clf()
        plt.close()

        if ref_ensemble and ref_energies:
            # Get the data for plotting of the ensemble
            ref_x = calculate_descriptor_for(
                x, ref_ensemble, *self._get_args_for(x))
            ref_y = calculate_descriptor_for(
                y, ref_ensemble, *self._get_args_for(y))
            image = get_energy_image(ref_x, ref_y, ref_energies, bins)
            # Label the plot
            plt.xlabel(x)
            plt.ylabel(y)
            plt.imshow(image.T, interpolation="nearest", origin='lower',
                       extent=[bins[0][0], bins[0][-1],
                               bins[1][0], bins[1][-1]],
                       aspect='auto')
            plt.xlim([bins[0][0], bins[0][-1]])
            plt.ylim([bins[1][0], bins[1][-1]])
            plt.colorbar()
            figname = "minEnergy_reference_{}_{}_{}.svg".format(
                self._cgs[0].name, x, y)
            plt.savefig(figname)
            log.info("Figure {} created".format(figname))
            plt.clf()
            plt.close()

        plt.set_cmap('jet')

    def view_2d_hist(self, ref_ensemble=None, x="rmsd_to_reference", y="rmsd_to_last", bins=None):
        """
        Plot a 2D histogram of the ensemble with respect to the given x and y axis,
        and visualize the results of clustering with DBSCAN.

        :param ref_ensemble: An ensemble or a list of cgs. Plotted as additional image.
        :param x: A STRING. The descriptor name used as x axis.
        :param y: A STRING. The descriptor name used as y axis.
        :param bins: Passed to matplotlib.pyplot.hist2d

        :returns: The bins used
        Saves the resulting plot as a svg in the current directory.
        """

        # Get the data for plotting of the ensemble
        data_x = self._get_descriptor(x)
        data_y = self._get_descriptor(y)

        if ref_ensemble:
            # Label the plot
            plt.xlabel(x)
            plt.ylabel(y)
            # Get the data for plotting of the ensemble
            ref_x = calculate_descriptor_for(
                x, ref_ensemble, *self._get_args_for(x))
            ref_y = calculate_descriptor_for(
                y, ref_ensemble, *self._get_args_for(y))
            if bins is None:
                # Calculate the bins, using reference and ensemble
                _, xedges, yedges = np.histogram2d(
                    list(ref_x) + list(data_x), list(ref_y) + list(data_y), bins=40, normed=True)
                bins = [xedges, yedges]
            # Plot the 2D histogram
            plt.hist2d(ref_x, ref_y, bins=bins, normed=True)
            plt.colorbar()
            figname = "hist2d_reference_{}_{}_{}.svg".format(
                ref_ensemble[0].name, x, y)
            plt.savefig(figname)
            log.info("Figure {} created".format(figname))
            plt.clf()
            plt.close()

        if bins is None:
            bins = 40
        # Label the plot
        plt.xlabel(x)
        plt.ylabel(y)
        # Plot the 2D histogram
        _, xedges, yedges, _ = plt.hist2d(
            data_x, data_y, bins=bins, normed=True)
        bins = [xedges, yedges]
        plt.colorbar()
        figname = "hist2d_{}_{}_{}.svg".format(self._cgs[0].name, x, y)
        plt.savefig(figname)
        log.info("Figure {} created".format(figname))
        plt.clf()
        plt.close()
        print("BINS", bins)
        return bins

    def view_2d_projection(self, ref_ensemble=None, x="rmsd_to_reference", y="rmsd_to_last", cluster=False, circular=False):
        """
        Plot a 2D projection of the ensemble to the given x and y axis,
        and visualize the results of clustering with DBSCAN.

        :param ref_ensemble: An ensemble or a list of cgs. Plotted as a background in the images.
        :param x: A STRING. The descriptor name used as x axis.
        :param y: A STRING. The descriptor name used as y axis.

        Saves the resulting plot as a svg in the current directory.
        """
        if circular:
            fig, ax = plt.subplots(1, subplot_kw=dict(projection='polar'))
        else:
            fig, ax = plt.subplots(1)

        # Label the plot
        ax.set_xlabel(x)
        ax.set_ylabel(y)

        # First, plot the background (reference ensemble)
        if ref_ensemble is not None:
            log.info("Reference ensemble given")
            ref_x = calculate_descriptor_for(
                x, ref_ensemble, *self._get_args_for(x))
            ref_y = calculate_descriptor_for(
                y, ref_ensemble, *self._get_args_for(y))

            log.info("Plotting reference")
            ax.plot(ref_x, ref_y, 's', markerfacecolor="green",
                    markeredgecolor='green', markersize=8)
        else:
            log.info("Reference ensemble missing")

        # Get the data for plotting of the ensemble
        data_x = self._get_descriptor(x)
        data_y = self._get_descriptor(y)

        if ref_ensemble is not None:
            # Without duplicates
            print("KS-Test without duplicates for {} : {}".format(x,
                                                                  scipy.stats.ks_2samp(data_x, ref_x)))
            print("KS-Test without duplicates for {} : {}".format(y,
                                                                  scipy.stats.ks_2samp(data_y, ref_y)))

            # With duplicates
            # Correctly account for duplicates
            full_x = [data_x[i] for i in self._cg_sequence]
            full_y = [data_y[i] for i in self._cg_sequence]
            print("KS-Test for {} : {}".format(x,
                                               scipy.stats.ks_2samp(full_x, ref_x)))
            print("KS-Test for {} : {}".format(y,
                                               scipy.stats.ks_2samp(full_y, ref_y)))

        if cluster:
            # In the background, plot lines to show the sampling trajectory
            ax.plot(data_x, data_y, '-', color="blue")

            # Then cluster all structures based on pairwise RMSD
            db = self._cluster_dbscan()
            labels = db.labels_
            core_samples_mask = np.zeros_like(db.labels_, dtype=bool)
            core_samples_mask[db.core_sample_indices_] = True
            unique_labels = set(labels)

            colors = plt.cm.Spectral(np.linspace(0, 1, len(unique_labels)))
            for k, col in zip(unique_labels, colors):
                if k == -1:
                    # Black used for noise.
                    col = 'k'
                class_member_mask = (labels == k)

                ax.plot(data_x[class_member_mask & core_samples_mask],
                        data_y[class_member_mask & core_samples_mask],
                        'o', markerfacecolor=col, markeredgecolor='k', markersize=6
                        )
                ax.plot(data_x[class_member_mask & ~core_samples_mask],
                        data_y[class_member_mask & ~core_samples_mask],
                        'o', markerfacecolor=col, markeredgecolor=col, markersize=1
                        )
        else:
            # In the background, plot lines to show the sampling trajectory
            ax.plot(data_x, data_y, '-o', color="blue")

        if self._reference_cg:
            ax.plot(calculate_descriptor_for(x, [self._reference_cg], *self._get_args_for(x)),
                    calculate_descriptor_for(
                        y, [self._reference_cg], *self._get_args_for(y)),
                    "x", color="red", markersize=12, label="reference")
        figname = "cluster_{}_{}_{}{}.svg".format(
            self._cgs[0].name, x, y, "circ" * circular)
        plt.savefig(figname)
        log.info("Figure {} created".format(figname))
        plt.clf()
        plt.close()

    def ensemble_pca(self, ref_ensemble=None, ref_first=True):
        data = prepare_pca_input(self._cgs)
        pca = PCA(n_components=2)
        if ref_ensemble:
            ref_data = prepare_pca_input(ref_ensemble)
            if ref_first:
                pca.fit(ref_data)
        if not ref_ensemble or not ref_first:
            pca.fit(data)
        reduced_data = pca.transform(data)
        if ref_ensemble:
            reduced_ref = pca.transform(ref_data)
            plt.scatter(reduced_ref[:, 0], reduced_ref[:, 1],
                        color="green", label="background")
        plt.scatter(reduced_data[:, 0], reduced_data[:,
                                                     1], color="blue", label="sampling")
        if self._reference_cg:
            data_true = prepare_pca_input([self._reference_cg])
            reduced_true = pca.transform(data_true)
            plt.scatter(reduced_true[:, 0], reduced_true[:,
                                                         1], color="red", label="reference")

        plt.xlabel("First principal component")
        plt.ylabel("Second principal component")
        figname = "pca_{}_rf{}.svg".format(self._cgs[0].name, ref_first)
        plt.savefig(figname)
        log.info("Figure {} created".format(figname))
        plt.clf()
        plt.close()

    def create_element_csv(self, outname):
        """
        :param outname: The outfile name to be written
        """
        data = defaultdict(list)
        for i, cg in enumerate(self._cgs):
            build_order = cg.traverse_graph()
            for elem in cg.mst:
                if elem[0] not in "mi":
                    continue
                line = cg.sampled[elem]
                # load angle_stats in direction of build order!
                for bo in build_order:
                    if bo[1] == elem:
                        stat = cg.get_bulge_angle_stats_core(
                            elem, (bo[0], bo[2]))
                stat.pdb_name = line[0]
                # Use correct multiplicity (self._cg has subsequent duplicates removed)
                for j in range(self._cg_sequence.count(i)):
                    data["cg_name"].append(cg.name)
                    data["key"].append(self._cg_rev_lookup[i][j])
                    data["elem_name"].append(elem)
                    data["stat_name"].append(stat.pdb_name)
                    data["u"].append(stat.u)
                    data["v"].append(stat.v)
                    data["angle"].append(stat.get_angle())
        df = pd.DataFrame(data)
        df.to_csv(outname)


def prepare_pca_input(cgs):
    data = []
    for cg in cgs:
        # reshape(-1) returns a flattened view
        data.append(ftuv.center_on_centroid(
            cg.get_ordered_stem_poss()).reshape(-1))
    return np.array(data)


def get_energy_image(data_x, data_y, energies, bins):
    image = np_nans([len(bins[0]) - 1, len(bins[1]) - 1])
    bins_x = np.digitize(data_x, bins[0])
    bins_y = np.digitize(data_y, bins[1])
    for i in range(len(data_x)):
        bx = bins_x[i] - 1
        by = bins_y[i] - 1
        if bx < 0:
            bx = 0
        if by < 0:
            by = 0
        if bx >= len(image):
            bx = len(image) - 1
        if by >= len(image[0]):
            by = len(image[0]) - 1
        if np.isnan(image[bx, by]):
            image[bx, by] = energies[i]
        else:
            image[bx, by] = min(energies[i], image[bx, by])
    return image


class DescriptorCalc(object):
    """
    Helper class to calculate descriptors of Coarse grained RNAs for an ensemble (or a list) of cgs.
    """

    @staticmethod
    def rmsd_to_stru(cgs, reference_cg):
        rmsd = np_nans(len(cgs))
        for i, cg in enumerate(cgs):
            rmsd[i] = ftms.cg_rmsd(cg, reference_cg)
        return rmsd

    @staticmethod
    def rog(cgs):
        rogs = np_nans(len(cgs))
        for i, cg in enumerate(cgs):
            rogs[i] = cg.radius_of_gyration()
        return rogs

    @staticmethod
    def anisotropy(cgs):
        ai = np_nans(len(cgs))
        for i, cg in enumerate(cgs):
            ai[i] = ftmd.anisotropy(cg.get_ordered_stem_poss())
        return ai

    @staticmethod
    def info_energy(cgs):
        e = np_nans(len(cgs))
        for i, cg in enumerate(cgs):
            try:
                e[i] = float(cg.infos["totalEnergy"][0].split()[0])
            except:
                e[i] = float("nan")
        return e

    @staticmethod
    def cg_distance(cgs, elem):
        d = []
        for cg in cgs:
            start, end = cg.coords[elem]
            d.append(ftuv.vec_distance(start, end))
        return d

    @staticmethod
    def stat_angle(cgs, elem):
        d = []
        for cg in cgs:
            angle = cg.get_stats(elem)[0].get_angle()
            d.append(angle)
        return d

    @staticmethod
    def cg_dist_difference(cgs, elem1, elem2):
        d = []
        for cg in cgs:
            start, end = cg.coords[elem1]
            d1 = ftuv.vec_distance(start, end)
            start, end = cg.coords[elem2]
            d2 = ftuv.vec_distance(start, end)
            d.append(d1 - d2)
        return d

    @staticmethod
    def cg_dist_sum(cgs, elem1, elem2):
        d = []
        for cg in cgs:
            start, end = cg.coords[elem1]
            d1 = ftuv.vec_distance(start, end)
            start, end = cg.coords[elem2]
            d2 = ftuv.vec_distance(start, end)
            d.append(d1 + d2)
        return d


valid_descriptors = {
    "rmsd_to_reference": DescriptorCalc.rmsd_to_stru,
    "rmsd_to_last": DescriptorCalc.rmsd_to_stru,
    "rog": DescriptorCalc.rog,
    "ROG": DescriptorCalc.rog,
    "anisotropy": DescriptorCalc.anisotropy,
    "info_energy": DescriptorCalc.info_energy
}


def calculate_descriptor_for(descriptor_name, cgs, *args):
    """Calculate a descriptor."""
    if descriptor_name.startswith("stat_angle"):
        elem = descriptor_name.split("_")[-1]
        return DescriptorCalc.stat_angle(cgs, elem)
    elif descriptor_name.startswith("cg_distance"):
        elem = descriptor_name.split("_")[-1]
        return DescriptorCalc.cg_distance(cgs, elem)
    elif descriptor_name.startswith("cg_dist_sum"):
        elem1, elem2 = descriptor_name.split("_")[-2:]
        return DescriptorCalc.cg_dist_sum(cgs, elem1, elem2)
    elif descriptor_name.startswith("cg_dist_difference"):
        elem1, elem2 = descriptor_name.split("_")[-2:]
        return DescriptorCalc.cg_dist_difference(cgs, elem1, elem2)
    elif descriptor_name not in valid_descriptors:
        raise ValueError("Unknown descriptor {}".format(descriptor_name))
    else:
        return valid_descriptors[descriptor_name](cgs, *args)



from __future__ import division, unicode_literals

import numpy as np


def serialize_vres(vpos):
    """
    Output carthesian coordinates for each virtual residue, ordered by nucleotide number and seperated by space.

    The returned coordinates are relative to the element's coordinate system.

    :param vpos: A dictionary `{nucleotide_number: coordinates}`.
                 `nucleotide_number` is an integer with 0 referring to the first nucleotide of the element.
                 `coordinates` is a 3-element 1D numpy array
                 `nucleotide_numbers` should be consecutive integers.
    :returns: A string with space-seperated numbers.
              Groups of three numbers are x,y,z coordinates of one residue.
    """
    out_str = ""
    for k, v in sorted(vpos.items(), key=lambda x: x[0]):
        out_str += "{:.8f} {:.8f} {:.8f} ".format(v[0], v[1], v[2])
    return out_str


def parse_vres(parts):
    """
    Go back from a splitted serialized representation to a dictionary used for vres.

    :param parts: A list of strings which hold floats. The length has to be a multiple of 3.
                  Like a list gegerated by `serialize_vres(vpos).split()`
    :returns: A dictionary {nucleotide_number: coordinated}. See param `vpos` of the
              function `serialize_vres`
    """
    vpos = {}
    vbase = {}
    vsugar = {}
    vbackbone = {}
    to_fill = vpos
    start = 0
    for to_fill in [vpos, vbase, vsugar, vbackbone]:
        for i, j in enumerate(range(start, len(parts), 3)):
            if parts[j].startswith("v"):
                start = j + 1
                break
            to_fill[i] = np.array(list(map(float, parts[j:j + 3])))

    return vpos, vbase, vsugar, vbackbone



# cython: profile=True
# distutils: language = c++
# distutils: sources = forgi/threedee/utilities/broken_ml_core.cpp

import numpy as np
cimport cython
cimport numpy as np
from libc.math cimport sqrt, cos, sin, acos
from libcpp.vector cimport vector


from forgi.threedee.utilities.broken_ml_core_py cimport Vect, _get_broken_ml_dev_core, testdot

@cython.boundscheck(False)
@cython.wraparound(False)  # turn off negative index wrapping for entire function
cdef double magnitude(double [:] vec) except -1:
    return sqrt(vec[0]*vec[0]+vec[1]*vec[1]+vec[2]*vec[2])

@cython.boundscheck(False)
@cython.wraparound(False)  # turn off negative index wrapping for entire function
cpdef np.ndarray[double, ndim=2] transposed_inverted(
                             np.ndarray[double, ndim=2] b):
    cdef double det = ( b[0,0]*(b[1,1]*b[2,2]-b[1,2]*b[2,1])
                      - b[0,1]*(b[1,0]*b[2,2]-b[1,2]*b[2,0])
                      + b[0,2]*(b[1,0]*b[2,1]-b[1,1]*b[2,0]))
    return np.array([
                        [(b[1,1]*b[2,2]-b[1,2]*b[2,1])/det, -(b[1,0]*b[2,2]-b[2,0]*b[1,2])/det, (b[1,0]*b[2,1]-b[2,0]*b[1,1])/det],
                        [-(b[0,1]*b[2,2]-b[2,1]*b[0,2])/det, (b[0,0]*b[2,2]-b[2,0]*b[0,2])/det, -(b[0,0]*b[2,1]-b[2,0]*b[0,1])/det],
                        [ (b[0,1]*b[1,2]-b[1,1]*b[0,2])/det, -(b[0,0]*b[1,2]-b[1,0]*b[0,2])/det, (b[0,0]*b[1,1]-b[1,0]*b[0,1])/det]
                    ])

@cython.boundscheck(False)
@cython.wraparound(False)  # turn off negative index wrapping for entire function
cpdef np.ndarray[double, ndim=2] create_orthonormal_basis(
                             double[:] vec1,
                             double[:] vec2): # Optimized
    cdef double ma, mb, mc
    cdef double ax,ay,az,bx,by,bz
    ma =magnitude(vec1)
    mb =magnitude(vec2)
    ax=vec1[0]/ma
    ay=vec1[1]/ma
    az=vec1[2]/ma
    bx=vec2[0]/mb
    by=vec2[1]/mb
    bz=vec2[2]/mb

    cdef np.ndarray[double, ndim=2] out = np.zeros((3,3))
    out[0,0]=ax
    out[0,1]=ay
    out[0,2]=az
    out[1,0]=bx
    out[1,1]=by
    out[1,2]=bz
    out[2,0]=ay*bz-az*by
    out[2,1]=az*bx-ax*bz
    out[2,2]=ax*by-ay*bx
    return out



def get_orig_coords(cg, broken_ml_name, fixed_stem_name):
    s1, s2 = cg.edges[broken_ml_name]
    if s1 == fixed_stem_name:
        orig_stem_name = s2
    elif s2 ==fixed_stem_name:
        orig_stem_name = s1
    else:
        raise ValueError("fixed stem {} is not attached to ml {} with "
                         "edges {}".format(fixed_stem_name, broken_ml_name, [s1,s2]))
    sides2 = cg.get_sides(orig_stem_name, broken_ml_name)
    cdef np.ndarray[double, ndim=1] orig_coords0 = cg.coords[orig_stem_name][sides2[0]]
    cdef np.ndarray[double, ndim=1] orig_coords1 = cg.coords[orig_stem_name][sides2[1]]
    cdef np.ndarray[double, ndim=1] orig_twist   =  cg.twists[orig_stem_name][sides2[0]]
    return orig_coords0, orig_coords1, orig_twist

def get_fixed_stemvec(cg, broken_ml_name, fixed_stem_name):
    sides = cg.get_sides(fixed_stem_name, broken_ml_name)
    cdef np.ndarray[double, ndim=1] fixed_s_vec = cg.coords.get_direction(fixed_stem_name)
    if sides[0]==0:
        fixed_s_vec = -fixed_s_vec
    cdef np.ndarray[double, ndim=1] fixed_s_pos = cg.coords[fixed_stem_name][sides[0]]
    cdef np.ndarray[double, ndim=1] s_twist = cg.twists[fixed_stem_name][sides[0]]

    return fixed_s_pos, fixed_s_vec, s_twist

def get_broken_ml_deviation(cg, broken_ml_name, fixed_stem_name, virtual_stat):
    cdef double[:] fixed_s_pos, fixed_s_vec, s_twist
    cdef double[:] orig_coords0, orig_coords1, orig_twist

    fixed_s_pos, fixed_s_vec, s_twist = get_fixed_stemvec(cg, broken_ml_name, fixed_stem_name)
    orig_coords0, orig_coords1, orig_twist = get_orig_coords(cg, broken_ml_name, fixed_stem_name)

    cdef Vect a =  _get_broken_ml_dev_core(
                Vect(fixed_s_pos[0], fixed_s_pos[1], fixed_s_pos[2]),
                Vect(fixed_s_vec[0], fixed_s_vec[1], fixed_s_vec[2]),
                Vect(orig_coords0[0], orig_coords0[1], orig_coords0[2]),
                Vect(orig_coords1[0], orig_coords1[1], orig_coords1[2]),
                Vect(orig_twist[0], orig_twist[1], orig_twist[2]),
                Vect(s_twist[0], s_twist[1], s_twist[2]),
                virtual_stat.r1, virtual_stat.u1, virtual_stat.v1,
                virtual_stat.u, virtual_stat.v, virtual_stat.t)
    return a.x, a.y, a.z



from forgi.graph.transform_graphs import _GCDummy, BGTransformer


class CGTransformer(BGTransformer):
    def condensed(self):
        """
        Return a condensed copy of the CoarseGrainRNA.

        See superclass docstring for definition of condensed.

        In the case of CoarseGrainRNA objects, the 3D coordinates are not
        affected by condensing the RNA. This means that stems which has multiple
        base-pairs have the same length as 1-bp stems in the condensed version.
        """
        new_cg = super(CGTransformer, self).condended()
        new_cg.coords = self.bg.coords
        new_cg.twists = self.bg.twists
        new_cg.sampled = self.bg.sampled
        new_cg.longrange = self.bg.longrange
        new_cg.chains = self.bg.chains
        return new_cg



"""
A module for dealing with modified residues.

For now the only functionality is to find out the corresponding
unmodified residue.
"""
from __future__ import print_function

from builtins import str
from builtins import object
try:
    from urllib.request import urlopen  # python3
except ImportError:
    from urllib import urlopen  # python2
import re
import logging
import warnings
import sys
from pprint import pprint


from ...graph import residue as fgr

#from .modified_res_lookup import RESIDUE_DICT

log = logging.getLogger(__name__)


def _parse_table_row(tr):
    tds = [td for td in tr.contents if td.name == "td"]
    if len(tds) == 2:
        key = tds[0].h3.text.strip()
        value = tds[1].text.strip()
        return key, value
    else:
        return None


def _html_to_info_dict(html):
    try:
        from bs4 import BeautifulSoup
    except ImportError:
        warnings.warn("Could not query PDBeChem, because BeautifulSoup is not installed"
                      "(install with `pip install forgi[pdbechem]` or `pip install forgi[all]`)")
        return {}
    with warnings.catch_warnings():
        # We don't care what xml parser beautiful soup uses
        warnings.simplefilter("ignore")
        parsed_html = BeautifulSoup(html)
    for h1 in parsed_html.body.find_all("h1"):
        if h1.text == "wwPDB Information":
            table = h1.parent.parent.parent  # (h1_>td->tr->tbody)
            rows = table.find_all("tr")
            info = {}
            for row in rows:
                data = _parse_table_row(row)
                if data:
                    info[data[0]] = data[1]
            return info
    return {}


def query_PDBeChem(three_letter_code):
    """
    Query the PDBeChem and return a dictionary with key-value pairs.

    Unfortunately, this information does not seem to be provided in any
    machine-readable format. We have to parse the html.
    """
    if len(three_letter_code) > 3:
        raise ValueError("Illegal 3-letter code")
    if not re.match("^[A-Z0-9][A-Z0-9][A-Z0-9]$", three_letter_code):
        raise ValueError("Illegal 3-letter code")
    html = urlopen(
        "http://www.ebi.ac.uk/pdbe-srv/pdbechem/chemicalCompound/show/{}".format(three_letter_code))
    residue_info = _html_to_info_dict(html)
    return residue_info


def change_residue_id(residue, new_id):
    chain = residue.parent
    if chain is not None and new_id in chain:
        raise ValueError("Cannot change id {old} to {new}. {new} already exists".format(
            old=residue.id, new=new_id))
    old_id = residue.id
    residue.id = new_id
    if chain is not None:
        try:
            del chain.child_dict[old_id]
        except KeyError:
            pass  # New version of biopython. Id is a property
        else:
            chain.child_dict[new_id] = residue


def to_4_letter_alphabeth(chain, query_db=False):
    '''
    Rename the modified residues so that they have the same
    names as unmodified residues.

    If a residue's name is unknown, query PDBeChem.

    If the residue is not a ribonucleotide, remove it from the chain.

    If it is a modified ribonucleotde, replace it by the unmodified "standard parent"


    :param chain: A Bio.PDB.Chain structure
    :param query_db: If true, query PDBeChem whenever a 3-letter code is unknown.
    :return: The same chain, but with only AUGC residues.
    '''
    import forgi.threedee.utilities.pdb as ftup
    modifications = {}
    i = 0
    while i < len(chain):
        r = chain.child_list[i]
        # Non-standard residues
        if r.resname.strip() == "I":
            # "I" has no standart parent (AUGC) to replace it with.
            warnings.warn("Inosine will be replaced by G")
            r.resname = "G"
            modifications[fgr.RESID(chain=chain.id, resid=(
                " ", r.id[1], r.id[2]))] = r.resname.strip()
        else:
            if r.resname.strip() not in ftup.RNA_RESIDUES:
                res_info = ModifiedResidueLookup(query_db)[r.resname]
                if not res_info:
                    # Unknown code. Remove residue
                    log.info(
                        "Detaching %s, because we do not understand the code %s.", r, r.resname)
                    chain.detach_child(r.id)
                    continue  # Continue with same i (now different residue)
                if res_info["Polymer type"] != "Ribonucleotide":
                    # DNA/ Amino Acid. Remove residue
                    chain.detach_child(r.id)
                    continue  # Continue with same i (now different residue)
                if res_info["Type description"] == "NON-POLYMER":
                    # e.g. GTP in 3DIR
                    log.warning(
                        "Detaching %s, because %s is classified as non-polymeric.", r, r.resname)
                    chain.detach_child(r.id)
                    continue  # Continue with same i (now different residue)
                if res_info["Standard parent"] not in ["A", "U", "G", "C"]:
                    log.warning("Detaching %s, because %s has standard parent '%s'.",
                                r, r.resname, res_info["Standard parent"])
                    chain.detach_child(r.id)
                    continue  # Continue with same i (now different residue)

                modifications[fgr.RESID(chain=chain.id, resid=(
                    " ", r.id[1], r.id[2]))] = r.resname.strip()
                r.resname = res_info["Standard parent"]

        # rename modified residues
        if r.id[0].strip():
            log.debug("Modified residue: %s", r.id)
            # The following was added because of 1NYI.pdb. It includes a single G-residue denoted as HETATOM in chain A.
            # It forms a base-pair, but is probably not connected to the backbone.
            # Instead of removing it/ in addition to removing it,
            # introducing cutpoint whereever the PDB Chain
            # contains gaps would be another option (TODO).
            # A plain AUGC as a HETATOM means it is a ligand.
            # 1Y26 has H_ADE instead of H_A
            if r.id[0].strip() in ["H_A", "H_U", "H_G", "H_C", "H_  G", "H_  C", "H_  A", "H_  U", "H_ADE", "H_CYT", "H_GUA", "H_URI"]:
                log.debug("Detaching %s, because it is a ligand", r.id)
                chain.detach_child(r.id)
                continue  # Continue with same i (now different residue)

            change_residue_id(r, (' ', r.id[1], r.id[2]))

        i += 1  # Go to next residue.
    return chain, modifications


class ModifiedResidueLookup(object):
    """
    Convenience wrapper to access modified_res_lookup.RESIDUE_DICT.
    If a key does not exist, query the database to add it."""

    def __init__(self, query_db=False):
        """
        :param query_db: If true, query PDBeChem whenever a 3-letter code is not understood.
        """
        from . import modified_res_lookup
        self._dict = modified_res_lookup.RESIDUE_DICT
        self._query_db= query_db
    def __getitem__(self, key):
        key = key.strip()
        if key not in self._dict:
            if not self._query_db:
                return None
            try:
                self._dict[key] = query_PDBeChem(key)
                log.debug("Successfully looked up %s", key)
            except:
                log.warning("Could not look-up modified residue key %s", key)
                self._dict[key] = None
        return self._dict[key]

    def clean_failed(self):
        for k, v in self._dict.items():
            if v is None:
                del self._dict[k]

    def __str__(self):
        return str(self._dict)


def dict_from_pdbs(filenames):
    # Just look at the SEQRES headers.
    ignore = set("AUGC")
    out_dict = {}
    for filename in filenames:
        with open(filename) as f:
            log.info("Opened file %s", filename)
            for line in f:
                if line.startswith("SEQRES"):
                    codes = line[20:].split()
                    for code in codes:
                        code = code.strip()
                        if code not in out_dict and code not in ignore:
                            try:
                                res_info = query_PDBeChem(code)
                                out_dict[code] = res_info
                            except (ValueError, LookupError) as e:
                                log.warning(
                                    "3-letter code '%s' not found: %s", code, e)
                                out_dict[code] = None
    return {k: v for k, v in out_dict.items()}


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    print("# Created by `python {}`".format(" ".join(sys.argv)))
    print("RESIDUE_DICT = ", end="")
    d = dict_from_pdbs(sys.argv[1:])
    for ion in ["MG", "NA", "K", "CL"]:
        if ion not in d:
            d[ion] = None
    pprint({k: v for k, v in d.items() if v is not None})



# cython: profile=True
import numpy as np
cimport cython
cimport numpy as np
from libc.math cimport sqrt, cos, sin, M_PI, acos

@cython.boundscheck(False)
@cython.wraparound(False)  # turn off negative index wrapping for entire function
cdef double magnitude(double [:] vec ) except -1:
    return sqrt(vec[0]*vec[0]+vec[1]*vec[1]+vec[2]*vec[2])


@cython.boundscheck(False)
@cython.wraparound(False)  # turn off negative index wrapping for entire function
cpdef np.ndarray[double, ndim=2] create_orthonormal_basis2(
                             np.ndarray[double, ndim=1] vec1,
                             np.ndarray[double, ndim=1] vec2): # Optimized
    cdef double ma, mb, mc
    cdef double ax,ay,az,bx,by,bz
    ma =magnitude(vec1)
    mb =magnitude(vec1)
    ax=vec1[0]/ma
    ay=vec1[1]/ma
    az=vec1[2]/ma
    bx=vec2[0]/mb
    by=vec2[1]/mb
    bz=vec2[2]/mb

    cdef np.ndarray[double, ndim=2] out = np.zeros((3,3))
    out[0,0]=ax
    out[0,1]=ay
    out[0,2]=az
    out[1,0]=bx
    out[1,1]=by
    out[1,2]=bz
    out[2,0]=ay*bz-az*by
    out[2,1]=az*bx-ax*bz
    out[2,2]=ax*by-ay*bx

    return out

@cython.boundscheck(False)
@cython.wraparound(False)  # turn off negative index wrapping for entire function
cpdef np.ndarray[double, ndim=2] create_orthonormal_basis3(
                             double[:] vec1,
                             double[:] vec2): # Optimized
    cdef double ma, mb, mc
    cdef double ax,ay,az,bx,by,bz
    ma =magnitude(vec1)
    mb =magnitude(vec1)
    ax=vec1[0]/ma
    ay=vec1[1]/ma
    az=vec1[2]/ma
    bx=vec2[0]/mb
    by=vec2[1]/mb
    bz=vec2[2]/mb

    cdef np.ndarray[double, ndim=2] out = np.zeros((3,3))
    out[0,0]=ax
    out[0,1]=ay
    out[0,2]=az
    out[1,0]=bx
    out[1,1]=by
    out[1,2]=bz
    out[2,0]=ay*bz-az*by
    out[2,1]=az*bx-ax*bz
    out[2,2]=ax*by-ay*bx

    return out

@cython.boundscheck(False)
@cython.wraparound(False)  # turn off negative index wrapping for entire function
cpdef np.ndarray[double, ndim=2] create_orthonormal_basis(
                        np.ndarray[double, ndim=1] vec1,
                        np.ndarray[double, ndim=1] vec2): # Slower
    vec1/=magnitude(vec1)
    vec2/=magnitude(vec2)
    cdef np.ndarray[double, ndim=1] vec3 = np.array([vec1[1]*vec2[2]-vec1[2]*vec2[1],  vec1[2]*vec2[0]-vec1[0]*vec2[2], vec1[0]*vec2[1]-vec1[1]*vec2[0] ])
    vec3/=magnitude(vec3)
    return np.array([vec1, vec2, vec3])



from __future__ import print_function, absolute_import, division

import os
import argparse
import warnings
import logging
import json

import forgi.utilities.commandline_utils as fuc
import forgi.threedee.classification._training.aminor_training as ftcta
log = logging.getLogger(__name__)


PDBIDS_LRSU = ["1HC8", "1MMS", "1MZP", "1VQO", "3jbu", "4IOA", "4LGT", "4QVI", "4v9p",
               "4w2g", "4wt8", "5D8H", "5imq", "5MMI", "5O60", "5V7Q", "3J79", "3J7Q", "5T2C"]
PDBIDS_SRSU = ["1G1X", "1I6U", "1KUQ", "2VQE", "4V19", "4v9o", "5lzd", "5mmm",
               "5o61", "5OOL", "5v93", "3J7A", "3J7P", "4UG0", "3JAM", "4P8Z",
               "4UG0", "4V5O", "4V88", "5FLX", "5LZS", "5OPT", "5XXU", "5XYI", "6AZ1"]

################################################################################
# Lets use this as a script
# Run as python -m forgi.threedee.classification._training.aminor_training
################################################################################


def main():
    parser = fuc.get_rna_input_parser("Train a classifier for A-Minior interactions.",
                                      nargs="+", rna_type="only_cg")
    parser.add_argument("--fr3d-result", type=str, required=True,
                        help="A file containing the FR3D output")
    parser.add_argument("--fr3d-query", type=str,
                        help="Add this string describing the FR3D query "
                             "as a comment to the trainingsdata-out file")
    parser.add_argument("--chain-id-mapping-dir", type=str,
                        help="If you use PDB-bundles, this directory "
                             "needs to hold all chain-id-mapping.txt files.")
    parser.add_argument("--trainingsdata-out", type=str,
                        default="forgi/threedee/data/aminor_geometries.csv",
                        help="File that will be written for the geometries "
                             "of interactions and non-interactions.")
    parser.add_argument("--model-params-out", type=str,
                        default="forgi/threedee/data/aminor_params.json",
                        help="File that will be written for the "
                             "model's hyper-parameters.")
    parser.add_argument("--test-set", type=str, help="':'-separated PDB-ids"
                                                     " for the test-set.")
    parser.add_argument("--train-set", type=str,
                        help="':'-separated PDB-ids for the train-set."
                             "Note: This is only used for cross-validation."
                             "The final model will be trained on all the data.")

    args = parser.parse_args()
    cgs, cg_filenames = fuc.cgs_from_args(args, rna_type="only_cg",
                                          return_filenames=True)
    ftcta.create_geometry_file(args.trainingsdata_out, cgs, cg_filenames,
                               args.fr3d_result, args.chain_id_mapping_dir,
                               args.fr3d_query)
    hyper_params = ftcta.tune_model(
        args.trainingsdata_out, args.train_set, args.test_set)
    with open(args.model_params_out, "w") as f:
        json.dump(hyper_params, f)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# knotted2nested.py

"""Script to remove pseudoknots from pseudoknotted structures.
Pseudoknot removal is the remaval of one or more base pairs from
a structural model (=list of base pairs), such that all the remaining
base pairs in the structure are nested.

All procedures are described in:
S. Smit, K. Rother, J. Heringa, and R. Knight
Manuscript in preparation.

FORMATS (-f and -F parameter)
The two input formats allowed are bpseq and ct.
    BPSEQ format (as on the Comparative RNA website):
    4 header lines
    sequence_position residue partner_position
    NOTE: If base is unpaired, partner is 0

    CT format:
    sequence_position residue seq_pos-1 seq_pos+1 partner_pos seq_pos
    NOTE: If base is unpaired, partner is 0

We support also VIENNA format as output. Standard dot-bracket string. Same
length as the corresponding sequence.

METHODS (-m parameter):
EC -- conflict elimination, most conflicting region gets removed first
EG -- conflict elimination, region with worst gain get removed first
IO -- incremental by order
IL -- incremental by length
IR -- incremental by range
OA -- all optimal solutions depending on optimization function and goal
OSR -- single optimal solution picked at random, also depend on function/goal
OSP -- single optimal solution picked by scoring certain properties, also
    depends on function/goal
NR -- Nussinov algorithm restricted to given list of base pairs

OPTIMIZATION FUNCTIONS (-o parameter)
BPS -- maximize the number of base pairs
HB -- maximize the number of hydrogen bonds in Watson-Crick and Wobble pairs
    (GC pair scores 3, AU and GU pairs score 2)

Author: Sandra Smit (S.Smit@few.vu.nl)

Revision History:

    File created on 20 Sep 2007.
    Slightly restructured by B. Thiel on 19 Sep 2016: main now takes filehandle
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from builtins import str
from builtins import object
from sys import exit
from .rna2d import Pairs
from optparse import OptionParser
from .bpseq import BpseqParser, BpseqParseError
from .ct_simple import MinimalCtParser, CtError
from .output import ct_output, bpseq_output, vienna_output
from .knots import conflict_elimination, ec, eg,\
    inc_order, inc_length, inc_range,\
    opt_all, opt_single_random, opt_single_property,\
    num_bps, hydrogen_bonds, nussinov_restricted

KNOWN_INPUT_FORMATS = {'bpseq': BpseqParser, 'ct': MinimalCtParser}
KNOWN_OUTPUT_FORMATS = {'bpseq': bpseq_output, 'ct': ct_output,
                        'vienna': vienna_output}
KNOWN_METHODS = {'EC': ec, 'EG': eg,
                 'IO': inc_order, 'IL': inc_length, 'IR': inc_range,
                 'OA': opt_all, 'OSR': opt_single_random, 'OSP': opt_single_property,
                 'NR': nussinov_restricted, }
KNOWN_OPT_METHODS = {'BPS': ('max', num_bps),
                     'HB': ('max', hydrogen_bonds)}

DEFAULT_INPUT_FORMAT = 'ct'
DEFAULT_OUTPUT_FORMAT = None
DEFAULT_METHOD = 'EG'
DEFAULT_OPT_METHOD = 'BPS'
DEFAULT_VERBOSE = False
DEFAULT_REMOVED = False


class KnottedStructure(object):
    """Lightweight object to hold a pseudoknotted structure
    """

    def __init__(self, Knotted, Seq=None, Header=None):
        """Initialize new KnottedStructure object

        Knotted -- Pairs object representing a (knotted) structure
        Seq -- RNA sequence
        Header -- list of header lines
        """
        self.Knotted = Knotted
        self.Seq = Seq
        self.Header = Header


class NestedStructure(object):
    """Lightweight object to hold a pseudoknot-free structure
    """

    def __init__(self, Nested, Seq=None, Header=None, Removed=None):
        """Initialize new NestedStructure object

        Nested -- Pairs object representing a nested structure
        Seq -- RNA sequence
        Header -- list of header lines
        Removed -- Pairs object containing the removed base pairs
        """
        self.Nested = Nested
        self.Seq = Seq
        self.Header = Header
        self.Removed = Removed


def parse_command_line_parameters():
    """Parse command line arguments"""

    # USAGE AND HELP INFO
    usage = '\n'.join(["Usage: python %prog [options] structure_file",
                       "See web or script documentation for an explanation" +
                       " on the formats and methods."])
    input_format_help =\
        "Input format. Options: [%s]. DEFAULT = ct."\
        % (', '.join(KNOWN_INPUT_FORMATS.keys()))
    output_format_help =\
        "Output format. Options: [%s]. DEFAULT = same as input format."\
        % (', '.join(KNOWN_OUTPUT_FORMATS.keys()))
    method_help = "Pseudoknot removal method. Options: [%s]. DEFAULT = EG."\
        % (', '.join(KNOWN_METHODS.keys()))
    opt_method_help = "Optimization method. Options: [%s]. DEFAULT = BPS."\
        % (', '.join(KNOWN_OPT_METHODS.keys()))

    result = OptionParser(usage=usage)

    # FLAGS
    result.add_option('-v', '--verbose', action='store_true', dest='verbose',
                      help="Run script with verbose output")
    result.add_option('-r', '--removed', action='store_true', dest='removed',
                      help="Report removed base pairs")
    # VLAUE PARAMETERS
    result.add_option('-f', '--input_format', action='store', type='string',
                      dest='input_format', metavar="FORMAT",
                      help=input_format_help)
    result.add_option('-F', '--output_format', action='store', type='string',
                      dest='output_format', metavar="FORMAT",
                      help=output_format_help)
    result.add_option('-m', '--method', action='store', type='string',
                      dest='method', metavar='METHOD',
                      help=method_help)
    result.add_option('-o', '--opt_method', action='store', type='string',
                      dest='opt_method', help=opt_method_help)

    # DEFAULT VALUES
    result.set_defaults(verbose=DEFAULT_VERBOSE)
    result.set_defaults(removed=DEFAULT_REMOVED)
    result.set_defaults(input_format=DEFAULT_INPUT_FORMAT)
    result.set_defaults(output_format=DEFAULT_OUTPUT_FORMAT)
    result.set_defaults(method=DEFAULT_METHOD)
    result.set_defaults(opt_method=DEFAULT_OPT_METHOD)

    return result.parse_args()


def k2n_main(input_filehandle, input_format=DEFAULT_INPUT_FORMAT,
             output_format=DEFAULT_OUTPUT_FORMAT, method=DEFAULT_METHOD,
             opt_method=DEFAULT_OPT_METHOD, verbose=DEFAULT_VERBOSE,
             removed=DEFAULT_REMOVED):
    """Return string of output containing nested structures (+commens)

    Use help function or see script documentation for an explanation
        of the parameters.
    """
    result = []  # list of lines with output

    # by default, output is in same format as input
    if output_format is None:
        output_format = input_format

    # =========================================================================
    # VALIDATE INPUT INFORMATION
    # =========================================================================
    # validate input, give feedback on invalid input
    if input_format not in KNOWN_INPUT_FORMATS:
        message = "'%s' is an invalid format. Valid options are [%s]"\
            % (format, ', '.join(KNOWN_INPUT_FORMATS.keys()))
        raise ValueError(message)
    if output_format not in KNOWN_OUTPUT_FORMATS:
        message = "'%s' is an invalid output format. Valid options are [%s]"\
            % (format, ', '.join(KNOWN_OUTPUT_FORMATS.keys()))
        raise ValueError(message)
    if method not in KNOWN_METHODS:
        message = "'%s' is an invalid method. Valid options are [%s]"\
            % (method, ', '.join(KNOWN_METHODS.keys()))
        raise ValueError(message)
    if opt_method not in KNOWN_OPT_METHODS:
        message =\
            "'%s' is an invalid optimization function. Valid options are [%s]"\
            % (opt_method, ', '.join(KNOWN_OPT_METHODS.keys()))
        raise ValueError(message)

    if verbose:
        result.append('INPUT FILE = %s;' % (input_file))
        result.append('FORMAT = %s; METHOD = %s;' % (input_format, method))

    # =========================================================================
    # COLLECT FILE INFORMATION
    # =========================================================================
    # For bpseq files, put info in a list, because it is always only one record
    input_parser = KNOWN_INPUT_FORMATS[input_format]
    input_data = []

    try:
        if input_format == 'bpseq':
            header, seq, pairs = input_parser(input_filehandle)
            import sys

            ks = KnottedStructure(pairs, Seq=seq, Header=header)
            input_data = [ks]
        else:
            for header, seq, pairs in input_parser(input_filehandle):
                input_data.append(KnottedStructure(Knotted=pairs, Seq=seq,
                                                   Header=header))
    except (BpseqParseError, CtError, ValueError) as e:
        message_lines = ["Cannot read input file.",
                         "Make sure the input file is in the specified format.",
                         "Internal error message:", "%s." % (str(e))]
        raise ValueError('\n'.join(message_lines))
    # =========================================================================
    # CREATE NESTED STRUCTURE(S)
    # =========================================================================
    if not input_data:
        raise ValueError("No input data found")

    pk_function = KNOWN_METHODS[method]

    output_data = []
    for knotted_struct in input_data:
        if method in ['OA', 'OSR', 'OSP']:
            if opt_method == 'HB':
                opt_goal, opt_wrapper =\
                    KNOWN_OPT_METHODS[opt_method]
                opt_function = opt_wrapper(seq)
            else:
                opt_goal, opt_function =\
                    KNOWN_OPT_METHODS[opt_method]
            if verbose:
                result.append('OPTIMIZATION GOAL = %s; FUNCTION = %s;'
                              % (opt_method, opt_goal))
        else:
            if verbose:
                result.append('OPTIMIZATION GOAL AND FUNCTION IGNORED')

        if method == 'OA':  # return value is a list of Pairs objects
            nested_result = pk_function(knotted_struct.Knotted,
                                        goal=opt_goal, scoring_function=opt_function,
                                        return_removed=removed)
            if removed:
                for nested, removed_pairs in nested_result:
                    ns = NestedStructure(Nested=nested, Seq=knotted_struct.Seq,
                                         Header=knotted_struct.Header, Removed=Pairs(removed_pairs))
                    output_data.append(ns)
                    if verbose:
                        result.append(
                            'NUMBER OF BASE PAIRS BEFORE: %s; AFTER: %s'
                            % (len(knotted_struct.Knotted), len(ns.Nested)))
            else:
                for nested in nested_result:
                    ns = NestedStructure(Nested=nested, Seq=knotted_struct.Seq,
                                         Header=knotted_struct.Header)
                    output_data.append(ns)
                    if verbose:
                        result.append(
                            'NUMBER OF BASE PAIRS BEFORE: %s; AFTER: %s'
                            % (len(knotted_struct.Knotted), len(ns.Nested)))
        elif method in ['OSR', 'OSP']:
            if removed:
                nested_pairs, removed_pairs = pk_function(pairs, goal=opt_goal,
                                                          scoring_function=opt_function, return_removed=removed)
                ns = NestedStructure(Nested=nested_pairs,
                                     Seq=knotted_struct.Seq,
                                     Header=knotted_struct.Header, Removed=Pairs(removed_pairs))
            else:
                nested_pairs = pk_function(pairs, goal=opt_goal,
                                           scoring_function=opt_function)
                ns = NestedStructure(Nested=nested_pairs,
                                     Seq=knotted_struct.Seq,
                                     Header=knotted_struct.Header)
            if verbose:
                result.append('NUMBER OF BASE PAIRS BEFORE: %s; AFTER: %s'
                              % (len(pairs), len(nested_pairs)))
            output_data.append(ns)
        else:  # return value is a single Pairs object
            if removed:
                nested_pairs, removed_pairs = pk_function(pairs,
                                                          return_removed=removed)
                ns = NestedStructure(Nested=nested_pairs,
                                     Seq=knotted_struct.Seq,
                                     Header=knotted_struct.Header, Removed=Pairs(removed_pairs))
            else:
                nested_pairs = pk_function(pairs)
                ns = NestedStructure(Nested=nested_pairs,
                                     Seq=knotted_struct.Seq,
                                     Header=knotted_struct.Header)
            if verbose:
                result.append('NUMBER OF BASE PAIRS BEFORE: %s; AFTER: %s'
                              % (len(pairs), len(nested_pairs)))
            output_data.append(ns)

    # =========================================================================
    # PRODUCE OUTPUT
    # =========================================================================

    output_function = KNOWN_OUTPUT_FORMATS[output_format]

    # return string with output
    for ns in output_data:
        result.append(output_function(seq=ns.Seq, pairs=ns.Nested,
                                      header_lines=ns.Header, removed=ns.Removed))
    return '\n'.join(result)


if __name__ == "__main__":

    # =========================================================================
    # PARSE COMMAND LINE PARAMETERS AND SEND INFO TO MAIN FUNCTION
    # =========================================================================

    opts, arg = parse_command_line_parameters()

    try:
        input_file = arg[0]
    except IndexError:
        raise ValueError("No input file specified")

    with open(input_filename, 'U') as input_filehandle:
        result = k2n_main(input_filehandle, input_format=opts.input_format,
                          output_format=opts.output_format, method=opts.method,
                          opt_method=opts.opt_method, verbose=opts.verbose, removed=opts.removed)
    print(result)



#!/usr/bin/env python
from __future__ import print_function

import logging
import os.path
import sys
import numpy as np

import forgi.threedee.model.similarity as ftms
import forgi.utilities.commandline_utils as fuc
import forgi.threedee.utilities.pdb as ftup

log = logging.getLogger(__name__)


def get_parser():
    parser = fuc.get_rna_input_parser("Compare RNA 3d structures based on RMSD, ACC"
                                      " or other measures. If more than 2 filenames are given, compare all files to the first.", "+", "3d", enable_logging=True)
    output_group = parser.add_argument_group("Controlling output",
                                             description="Control, based on which measure the two "
                                             "structures will be compared")
    output_group.add_argument("--acc", help="Compare based on the Adjacency correlation coefficient"
                                            " ACC", action="store_true")
    output_group.add_argument(
        "--rmsd", help="Compare based on CG-RMSD", action="store_true")
    output_group.add_argument(
        "--stem-rmsd", help="Compare based on coarse grained stem RMSD", action="store_true")
    output_group.add_argument(
        "--pdb-rmsd", help="Compare based on PDB-RMSD", action="store_true")
    return parser


def main(args):
    with fuc.hide_traceback():
        cgs = fuc.cgs_from_args(
            args, rna_type="3d", enable_logging=True)

        if len(cgs)<2:
            raise ValueError("At least 2 RNA structures required for comparison")
        cg1=cgs[0]
        if not (args.acc or args.rmsd or args.pdb_rmsd or args.stem_rmsd):
            showall = True
        else:
            showall = False
        for cg2 in cgs[1:]:
            if showall or args.acc:
                if cg1.defines != cg2.defines:
                    if args.acc:
                        print(
                           "Cannot compare two 3d structures that do not correspond to the same RNA.")
                        sys.exit(1)
                else:
                    adj = ftms.AdjacencyCorrelation(cg1)
                    print("ACC:\t{:.3f}".format(ftms.mcc(adj.evaluate(cg2))))
            if showall or args.rmsd:
                print("RMSD:\t{:.3f}".format(ftms.cg_rmsd(cg1, cg2)))
            if showall or args.stem_rmsd:
                try:
                    print("Stem:\t{:.3f}".format(ftms.cg_stem_rmsd(cg1, cg2)))
                except ftms.Incompareable:
                    if args.stem_rmsd:
                        raise
                    # Otherwise do nothing
            if showall or args.pdb_rmsd:
                if not pdb_rmsd(cg1, cg2):
                    # If --pdb-rmsd was not given, just don't print it.
                    # If it was given, we exit with non-zero exit status.
                    if args.pdb_rmsd:
                        print("Cannot calculate PDB-RMSD: "
                              "The two files do not contain the same chains.")

def pdb_rmsd(cg1, cg2):
    if len(cg1.chains)==1==len(cg2.chains):
        reslist1 = []
        reslist2 = []
        chain1, = cg1.chains.values()
        chain2, = cg2.chains.values()
        resnames1=[ r.resname.strip() for r in chain1 ]
        resnames2=[ r.resname.strip() for r in chain2 ]
        log.info("Resname lists are %s, %s", resnames1, resnames2)
        if resnames1 == resnames2:
            for r in chain1:
                if r.resname.strip() in ftup.RNA_RESIDUES:
                    reslist1.append(r)
            for r in chain2:
                if r.resname.strip() in ftup.RNA_RESIDUES:
                    reslist2.append(r)
            print("PDB-RMSD (single chain):\t{:.3f}".format(
                      ftup.residuelist_rmsd(reslist1, reslist2, sidechains=True)[1]))
            return True

        else:
            log.info("Resname lists are different")
            reslist1, reslist2 = common_reslists(chain1, chain2)
            print("PDB-RMSD (single chain):\t{:.3f}".format(
                      ftup.residuelist_rmsd(reslist1, reslist2, sidechains=True)[1]))
            return True
    else:
        log.info("Chain lengths are %s and %s",len(cg1.chains), len(cg2.chains) )
        common_chains = set(cg1.chains.keys()) & set(cg2.chains.keys())
        reslist1 = []
        reslist2 = []
        for chain in common_chains:
            ext1, ext2 = common_reslists(cg1.chains[chain], cg2.chains[chain])
            reslist1.extend(ext1)
            reslist2.extend(ext2)
        if common_chains:
            print("PDB-RMSD (chains {}):\t{:.3f}".format("-".join(common_chains),
                  ftup.residuelist_rmsd(reslist1, reslist2, sidechains=True)[1]))
            return True
        return False


def common_reslists(chain1, chain2):
    reslist1 = []
    reslist2 = []
    start1 = chain1.get_list()[0].id[1]
    start2 = chain2.get_list()[0].id[1]
    offset=start2-start1
    mismatches=float("inf")
    for offs in [0, offset]:
        rl1, rl2, m = _common_reslist_with_offset(chain1, chain2, offs)
        if m<mismatches:
            mismatches=m
            reslist1, reslist2 = rl1, rl2
    if mismatches:
        log.warning("There were %s point mutations between the two chains", mismatches)
    if mismatches>8:
        raise ValueError("Could not find mapping between chains!")
    return reslist1, reslist2

def _common_reslist_with_offset(chain1, chain2, offset):
    log.info("Trying mapping with offset %s", offset)
    reslist1 = []
    reslist2 = []
    for cr2 in chain2.get_list():
        mapped_id = cr2.id[0], cr2.id[1]-offset, cr2.id[2]
        if mapped_id not in chain1:
            log.info("Residue %s (%s) [->%s] not in chain1", cr2.id, cr2.resname.strip(), mapped_id)
            log.info("Res %s has atoms %s", mapped_id, cr2.get_list())
    mismatch=0
    for cr1 in chain1.get_list():
        mapped_id = cr1.id[0], cr1.id[1]+offset, cr1.id[2]
        cr1_name = cr1.resname.strip()
        try:
            cr2 = chain2[mapped_id]
        except KeyError as e:
            log.info("Residue %s (%s) [->%s] not in chain2", cr1.id, cr1_name, mapped_id)
            continue
        cr2_name = cr2.resname.strip()
        if cr1_name != cr2_name:
            mismatch+=1
            log.info("Residue %s [->%s](chains %s, %s) has different name in the "
                        "two chains: %s!=%s", cr1.id, mapped_id, chain1.id, chain2.id, cr1_name, cr2_name)
        if cr1_name in ftup.RNA_RESIDUES and cr2_name in ftup.RNA_RESIDUES:
            reslist1.append(cr1)
            reslist2.append(cr2)

    return reslist1, reslist2, mismatch



parser = get_parser()
if __name__ == "__main__":
    args = parser.parse_args()
    main(args)



"""
This modules contains classifiers which classify a given coarse grained
geometry as a certain type of interaction.

An example would be classifying relative element orientations as A-minor
interaction or stacking interaction.

Classifiers from this module are compatible with scikit-learn.
"""



#!/usr/bin/python
from builtins import input

import argparse
import json
import sys


import forgi.config

parser = argparse.ArgumentParser(
    "Create/ update configuration files for forgi.")

if __name__ == "__main__":
    config = {}
    for filename in forgi.config.iter_configfiles():
        try:
            with open(filename) as f:
                conf = json.load(f)
        except (OSError, IOError):
            print("No config file exists yet at {}".format(filename))
        else:
            if conf:
                print("The following values are set in {}:".format(filename))
                for k, v in conf.items():
                    print("   {:15s}: {}".format(k, repr(v)))
                config.update(conf)
            else:
                print("This config file is empty")
    for filename in forgi.config.iter_configfiles():
        update = input("Update the file {}? (y/N)".format(filename))
        if update in ["y", "Y"]:
            with open(filename) as f:
                conf = json.load(f)
            for key in forgi.config.ALLOWED_KEY_VALUES:
                if key in config:
                    print("After parsing all config files, {} is set to {}".format(
                        key, config[key]))
                    if key in conf:
                        print("In the file you are editing, {} is set to {}".format(
                            key, conf[key]))
                    else:
                        print(
                            "In the file you are editing, {} is not set".format(key))
                else:
                    print("Currently, {} is not set in any config file".format(key))
                while True:
                    new_value = input("Choose value for {} ({}, empty input leave it unchanged, 'X' to delete the entry from the config file)".format(
                        key, "/".join(forgi.config.ALLOWED_KEY_VALUES[key])))
                    if not new_value:
                        break
                    if new_value in forgi.config.ALLOWED_KEY_VALUES[key]:
                        conf[key] = new_value
                        break
                    if new_value == "X":
                        del conf[key]
                        break
                    # Else next loop execution
            with open(filename, "w") as f:
                json.dump(conf, f)
                print("Configuration file {} updated".format(filename))



from __future__ import print_function, absolute_import, division

import os
import argparse
try:
    from collections.abc import Set
except ImportError:
    from collections import Set
from collections import namedtuple, defaultdict
import warnings
import logging
import time
import subprocess

import numpy as np
from sklearn.model_selection import GridSearchCV, train_test_split
from sklearn.metrics import confusion_matrix, f1_score

import pandas as pd

from logging_exceptions import log_to_exception
from .. import aminor as ftca
import forgi.utilities.stuff as fus
from ._parse_utils import ChainIdMappingParser
import forgi.graph.residue as fgr
import forgi.graph.bulge_graph as fgb


try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError

log = logging.getLogger(__name__)


PDBIDS_LRSU = ["1HC8", "1MMS", "1MZP", "1VQO", "3jbu", "4IOA", "4LGT", "4QVI", "4v9p",
               "4w2g", "4wt8", "5D8H", "5imq", "5MMI", "5O60", "5V7Q", "3J79", "3J7Q", "5T2C"]
PDBIDS_SRSU = ["1G1X", "1I6U", "1KUQ", "2VQE", "4V19", "4v9o", "5lzd", "5mmm",
               "5o61", "5OOL", "5v93", "3J7A", "3J7P", "4UG0", "3JAM", "4P8Z",
               "4UG0", "4V5O", "4V88", "5FLX", "5LZS", "5OPT", "5XXU", "5XYI", "6AZ1"]


def create_geometry_file(outfile, cgs, cg_filenames, fr3d_result,
                         chain_id_mapping_dir, fr3d_query=""):
    ame, non_ame = from_fr3d_to_orientation(
        cgs, fr3d_result, chain_id_mapping_dir)
    # First of all, print the ame_geometries.csv file!
    with open(outfile, "w") as file_:
        print("# Generated on " + time.strftime("%d %b %Y %H:%M:%S %Z"), file=file_)
        print("# Generated from the following files: ", file=file_)
        for fn in cg_filenames:
            print("#   {}".format(fn), file=file_)
        print("# Version: {}".format(fus.get_version_string().strip()), file=file_)
        print("# fr3d_result = {}".format(fr3d_result), file=file_)
        if fr3d_query:
            print("# fr3d_query:", file=file_)
            for line in fr3d_query.splitlines():
                line = line.strip()
                print("#    " + line, file=file_)
        print("# cutoff_dist = {} A".format(ftca.CUTOFFDIST), file=file_)
        # HEADER
        print ("pdb_id loop_type dist angle1 angle2 is_interaction "
               "loop_sequence interaction score annotation loop_flexibility", file=file_)
        for entry in ame:
            print("{pdb_id} {loop_type} {dist} {angle1} "
                  "{angle2} {is_interaction} {loop_sequence} "
                  "{loop_name}-{stem_name} {score} "
                  "\"{annotation}\" {loop_flexibility}".format(is_interaction=True,
                                                               **entry._asdict()), file=file_)
        for entry in non_ame:
            print("{pdb_id} {loop_type} {dist} {angle1} "
                  "{angle2} {is_interaction} {loop_sequence} "
                  "{loop_name}-{stem_name} {score} "
                  "\"{annotation}\" {loop_flexibility}".format(is_interaction=False,
                                                               **entry._asdict()), file=file_)


def tune_model(geometry_file, trainset, testset):
    if not trainset:
        trainset = PDBIDS_LRSU
    if not testset:
        testset = PDBIDS_SRSU
    df = pd.read_csv(geometry_file, comment="#", sep=" ")
    all_params = {}
    for loop_type in "hi":
        pI = calculate_pI(df, loop_type)
        # sum(labels)/(sum(labels)+len(df[mask_non_fred]))
        log.info("PI %s", pI)
        all_params[loop_type] = find_hyperparameters(loop_type, df,
                                                     pI, trainset, testset)
    return all_params

################################################################################
# Validate Model
################################################################################


def calculate_pI(df, loop_type):
    mask_ame, mask_non_ame, mask_non_fred = ftca._get_masks(df, loop_type)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        df["loop_name"] = df.pdb_id + \
            df.interaction.str.split("-").apply(lambda x: x[0])
    negative = len(df[mask_non_fred].drop_duplicates(["loop_name"]))
    positive = len(df[mask_ame].drop_duplicates("loop_name"))
    return positive / (negative + positive)


def mask_data(df, pdbids):
    mask = df.pdb_id.str.upper().str.contains("testestest")
    for pdb_id in pdbids:
        mask |= df.pdb_id.str.upper().str.contains(pdb_id.upper())
    return df[mask]


def find_hyperparameters(loop_type, df, pI, trainset, testset):
    log.info("Now searching for bandwidth")
    df_train = mask_data(df, trainset)
    df_test = mask_data(df, testset)
    if len(df_train) == 0 or len(df_test) == 0:
        raise ValueError("Either trainings- or testset are empty.")
    X_train, y_train = ftca.df_to_data_labels(df_train, loop_type)
    X_test, y_test = ftca.df_to_data_labels(df_test, loop_type)
    log.info("Length trainingsset: %s, length testset: %s",
             len(X_train), len(X_test))

    # We use the gaussian kernel to avoid certain artefacts of the linear
    # one and symmetric due to geometric considerations.
    clf = ftca.AMinorClassifier(p_I=pI, kernel="gaussian", symmetric=True)
    clf.fit(X_train, y_train)
    scores = []
    bandwidths = np.linspace(0.01, 1.0, 25)
    for bandwidth in bandwidths:
        log.debug("bandwidth %s", bandwidth)
        clf.set_params(bandwidth=bandwidth)
        score = clf.score(X_test, y_test)
        log.debug("score is %s", score)
        scores.append(score)
    best_i = scores.index(max(scores))
    best_bandwidth = bandwidths[best_i]
    clf.set_params(bandwidth=best_bandwidth)
    y_pred = clf.predict(X_test)
    tn, fp, fn, tp = confusion_matrix(y_test, y_pred).ravel()
    specificity = tn / (tn + fp)
    sensitivity = tp / (tp + fn)
    log.info("Classification report for loop type %s (using CV, bw=%s)\n:"
             "tp: %s, fp: %s, tn: %s, fn: %s, sens: %s, spec; %s",
             loop_type, best_bandwidth, tp, fp, tn, fn, sensitivity, specificity)
    if sensitivity < 0.8 or specificity < 0.8:
        warnings.warn("The crossvalidation score for loop tpype {} is "
                      "worse than expected: sensitivity: {},"
                      " specificity: {}.".format(loop_type, sensitivity, specificity))
    if sensitivity < 0.5 or specificity < 0.7:
        raise RuntimeError("Something went wrong with the training of the "
                           "classifier. The crossvalidation score is "
                           "terrible: sensitivity: {},"
                           " specificity: {}.".format(sensitivity, specificity))
    return clf.get_params()


################################################################################
# From FRED to TRAININGSDATA
################################################################################
_AMGeometry = namedtuple("AMGeometry",
                         ["pdb_id", "loop_name", "stem_name", "dist", "angle1",
                          "angle2", "loop_sequence", "score", "annotation",
                          "loop_flexibility"])


class AMGeometry(_AMGeometry):
    def _asdict(self):
        d = super(AMGeometry, self)._asdict()
        d["loop_type"] = self.loop_type
        return d

    @property
    def loop_type(self):
        return self.loop_name[0]


class AmeGeometrySet(Set):
    """
    Like a set, but use only pdb_id, loop_name and stem_name for
    equality, not the hash.

    If multiple geometries with the same key are added, use the lowest-scoring.
    """

    def __init__(self):
        self.geometries = {}

    @staticmethod
    def _geomerty_to_key(geo):
        return (geo.pdb_id, geo.loop_name, geo.stem_name)

    def add(self, geometry):
        key = self._geomerty_to_key(geometry)
        if key in self.geometries:
            # Smaller score is better
            if geometry.score > self.geometries[key].score:
                log.info("Duplicate geometry: %s has worse score "
                         "than %s", geometry, self.geometries[key])
                return
            else:
                log.info("Duplicate geometry: "
                         "%s replacing %s", geometry, self.geometries[key])
        self.geometries[key] = geometry

    def __contains__(self, geo):
        key = self._geomerty_to_key(geo)
        return key in self.geometries

    def __iter__(self):
        for g in self.geometries.values():
            yield g

    def __len__(self):
        return len(self.geometries)


def _chains_from_fr3d_field(value, pdb_id, mapping_directory):
    if len(value) == 3:
        chains = value
    elif len(value) == 6:
        chains = [value[:2], value[2:4], value[4:]]
        # Asterix as a special character for pdbs with length 1-chain-ids in bundles
        chains = list(map(lambda x: x.replace('*', ''), chains))
    else:
        raise ValueError(
            "PDB-id {}: Expecting either 3 or 6 letters for chain, found '{}'.".format(pdb_id, value))
    try:
        chain_id_mapping_file = os.path.join(
            mapping_directory, pdb_id.lower() + "-chain-id-mapping.txt")
        mapping = ChainIdMappingParser().parse(chain_id_mapping_file).mmcif2bundle
    except FileNotFoundError:
        return chains
    else:
        return [mapping[c] for c in chains]


def _parse_fred_line(line, all_cgs, current_annotation, mapping_directory):
    parts = line.split()
    if len(parts) < 10:
        return
    if parts[0] == "Filename":
        return
    pdb_id = parts[0]

    if parts[0] in all_cgs:
        chains = _chains_from_fr3d_field(parts[8], parts[0], mapping_directory)
        if isinstance(chains, list):  # A pdb bundle:
            if not all(c[0] == chains[0][0] for c in chains):
                log.info("Found an Interaction between multiple "
                         "files in pdb-bundle: {} {}. "
                         "Ignoring it!".format(parts[0], [c[0] for c in chains]))
                return
            pdb_name = chains[0][0].split(".")[0]
            chains = [c[1] for c in chains]
            log.debug(
                "For bundle: pdb_name = %s, converted-chains = %s", pdb_name, chains)
        else:
            pdb_name = parts[0]
        cgs = all_cgs[parts[0]]
        if len(cgs) == 1:
            cg, = cgs
        else:
            # First, filter by pdb-bundle file.
            cgs = [cg for cg in cgs if pdb_name in cg.name]
            log.debug("cgs for pbd-name %s are %s",
                      pdb_name, [c.name for c in cgs])
            # Then for multiple chains, search based on resid.
            a_res = _safe_resid_from_chain_res(
                chain=chains[0], residue=parts[3])
            if a_res is None:
                warnings.warn("Cannot create resid for {}".format(chains[0]))
                return
            for cg in cgs:
                # Find the first matching cg.
                if a_res in cg.seq._seqids:
                    break
            else:
                warnings.warn(
                    "No CoarseGrainRNA found among those with PDB-ID {} that contains resid {}".format(parts[0], a_res))
                return
    else:
        warnings.warn(
            "No CoarseGrainRNA found for FR3D annotation with PDB-ID {}. Possible PDB-IDs are {}".format(parts[0], all_cgs.keys()))
        return
    log.debug("FR3D line refers to cg %s", cg.name)
    # We now have the CoarseGrainRNA. Get the interacting cg-elements.
    nums = []
    for i in range(3):
        seq_id = _safe_resid_from_chain_res(
            chain=chains[i], residue=parts[3 + 2 * i])
        if seq_id is None:
            warnings.warn("Cannot create resid for {}".format(chains[i]))
            return
        try:
            nums.append(cg.seq_id_to_pos(seq_id))
        except ValueError:
            if seq_id.chain not in cg.chains:
                warnings.warn("Chain {!r} is not part of the cg {}. Required chains: {}. Available "
                              "cgs: {}".format(seq_id.chain, cg.name, chains,
                                               list(map(lambda x: x.name, all_cgs[parts[0]]))))
                return
            else:
                log.error("%s %s", cg.seq, type(cg.seq))
                log.error("All seq_ids=%s", cg.seq._seqids)
                raise
    nodes = list(map(lambda x: cg.get_node_from_residue_num(x), nums))
    if nodes[1] != nodes[2] or nodes[1][0] != "s":
        log.info("Parse_fred: No canonical stem: {} != {}.".format(
            nodes[1], nodes[2]))
        return  # only look at canonical stems.
    if nodes[0][0] == "s":
        log.info(
            "Parse_fred: Stem-Stem A-Minor not allowed: {} -> {}.".format(nodes[0], nodes[1]))
        return  # No A-Minor between two stems.
    if nodes[0] in cg.edges[nodes[1]]:
        log.info(
            "Parse_fred:  A-Minor between adjacent elements not allowed: {} -> {}.".format(nodes[0], nodes[1]))
        return  # Only look at not connected elements
    dist, angle1, angle2 = ftca.get_relative_orientation(
        cg, nodes[0], nodes[1])
    if np.isnan(angle1 + angle2 + dist):
        warnings.warn(
            "Cannot get relative orientation. Zero-length element {}".format(nodes[0]))
        return
    if nodes[0][0] == "i":
        flexibility = ftca.get_loop_flexibility(cg, nodes[0])
    else:
        flexibility = 1
    return (AMGeometry(cg.name, nodes[0], nodes[1], dist, angle1, angle2,
                       "&".join(cg.get_define_seq_str(nodes[0])),
                       float(parts[1]), current_annotation,
                       flexibility))


def _safe_resid_from_chain_res(chain, residue):
    try:
        return fgr.resid_from_str(str("{}:{}".format(chain, residue)))
    except ValueError as e:
        if residue.isdigit():
            with log_to_exception(log, e):
                log.error("Chain is '{}', res is '{}'".format(chain, residue))
            raise
        else:
            warnings.warn("Illegal residue number: '{}'.".format(residue))
            return


def parse_fred(cutoff_dist, all_cgs, fr3d_out, chain_id_mapping_dir):
    """
    Used by generate_target_distribution.

    :param all_cgs: A dictionary PDB_id (4-letter): [CoarseGrainRNA, ...]
    :param fr3d_out: A file opened for reading.

    :returns: A set of AMinor Geometries and a list of PDB-IDs, for which at
              least one line did not lead to an Aminor Geometry
    """
    with warnings.catch_warnings():
        warnings.simplefilter('always', UserWarning, append=True)
        geometries = AmeGeometrySet()
        skipped = 0
        #: What type of AMinor interactions.
        #: Comments in the form "# AMinor 0" have to be added manually
        #: when copying the FR3D-output to a file.They should preced the
        #: actual FR3D-output to which they apply.
        current_annotation = "?"
        for line in fr3d_out:
            line = line.strip()
            if not line:  # Empty line
                continue
            if line.startswith("# AMinor"):
                current_annotation = line[9:]
            elif line.startswith("#"):
                current_annotation = "?"
            log.debug("Line '%s'.read", line)
            geometry = _parse_fred_line(
                line, all_cgs, current_annotation, chain_id_mapping_dir)
            if geometry is None:
                if not (line.startswith("Filename") or line.startswith("#")):
                    skipped += 1
                    log.info("Skipping line {!r}".format(line))
            elif geometry.dist > cutoff_dist:
                log.debug("Skipping because of %f > %f (=cutoff dist): %r",
                          geometry.dist, cutoff_dist, line)
            elif "A" not in geometry.loop_sequence:
                warnings.warn("No adenine in loop %r for line %r",
                              geometry.loop_name, line)
            else:
                geometries.add(geometry)
    return geometries, skipped


def from_fr3d_to_orientation(cgs, fr3d_outfile, chain_id_mapping_dir):
    """
    Generate the trainings data for the AMinor classification
    from a list of cgs and the FR3D results of a query against
    the corresponding pdbs.

    ..note::

        The FR3D query needs to contain 3 nucleotides where the first is the Adenine and
        the second and third form a basepair in a stem.

    It needs FR3D output like this::

        Filename Discrepancy       1       2       3 Cha   1-2   1-3   2-3 Con   1-2   1-3   2-3   1-2   1-3   2-1   2-3   3-1   3-2   1-2   1-3   2-1   2-3   3-1   3-2   1-2   1-3   2-3
            1S72      0.0340  A  104  A  957  U 1009 900 ----  ----   cWW  AAA  1909  1885    24                                                                             -     -     -

    :param cgs: A list of CoarseGrainRNAs
                .. note::

                    the first 4 characters of the CoarseGrainRNA's name must match the
                    pdb-id as reported by FR3D.

    :param fr3d_out: A file containing the annotations found by FR3D.
    :param orientation_outfile: Where to store the list of relative orientations of
                                Adenines and the Stems in A-Minor interactions and
                                in the background. Relative to "fess/"
    :param fr3d_query: Describe the FR3D-query that you used in a string.
                       It will be added as comment to the output file.
    :param chain_id_mapping_dir: A dictionary containing the
                      chain-id-mapping files for pdb-bundles.
    """
    # Create lookup-dictionary from cgs.
    all_cgs = defaultdict(list)
    for cg in cgs:
        all_cgs[cg.name[:4].upper()].append(cg)
    # Read the FR3D output
    with open(fr3d_outfile) as f:
        aminor_geometries, num_skipped = parse_fred(
            ftca.CUTOFFDIST, all_cgs, f, chain_id_mapping_dir)
    log.info("%d entries skipped during FR3D parsing, %s entries retained",
             num_skipped, len(aminor_geometries))
    if len(aminor_geometries) == 0:
        raise ValueError(
            "No A-Minor geometries found. Is the FR3D output file correct?")
    non_ame_geometries = _enumerate_background_geometries(
        all_cgs, ftca.CUTOFFDIST, aminor_geometries)
    return aminor_geometries, non_ame_geometries


def _enumerate_background_geometries(all_cgs, cutoff_dist, aminor_geometries):
    """
    :param all_cgs: A dictionary {PDBID: [ cg1, cg2, ...]}
    """
    non_ame_geometries = set()
    for pdb_id, curr_cgs in all_cgs.items():
        for cg in curr_cgs:
            try:
                for loop in cg.defines:
                    if loop[0] == "s":
                        continue
                    if loop in cg.incomplete_elements or loop in cg.interacting_elements:
                        continue
                    for stem in cg.stem_iterator():
                        if loop in cg.edges[stem]:
                            continue
                        if stem in cg.incomplete_elements or stem in cg.interacting_elements:
                            continue
                        dist, angle1, angle2 = ftca.get_relative_orientation(
                            cg, loop, stem)
                        if loop[0] == "i":
                            flexibility = ftca.get_loop_flexibility(cg, loop)
                        else:
                            flexibility = 1
                        if not np.isnan(dist + angle1 + angle2) and dist <= cutoff_dist:
                            geometry = AMGeometry(cg.name, loop, stem, dist,
                                                  angle1, angle2,
                                                  "&".join(
                                                      cg.get_define_seq_str(loop)),
                                                  1000, "no_interaction",
                                                  flexibility)
                            if geometry in aminor_geometries:
                                log.info(
                                    "Geometry %s is in aminor_geometries", geometry)
                            else:
                                non_ame_geometries.add(geometry)
            except BaseException as e:
                with log_to_exception(log, e):
                    log.error(
                        "An Error occurred during processing of cg: %s", cg.name)
                raise
    log.error("%s non_ame geometries found", len(non_ame_geometries))
    return non_ame_geometries



#!/usr/bin/env python
# bpseq.py

"""Provides parser for bpseq files downloaded from Gutell's comparative
RNA website (CRW): http://www.rna.icmb.utexas.edu/

Author: Sandra Smit (Sandra.Smit@colorado.edu)

The file format is:

Filename: d.16.b.E.coli.bpseq
Organism: Escherichia coli
Accession Number: J01695
Citation and related information available at http://www.rna.icmb.utexas.edu
1 A 0
2 A 0
3 A 0
4 U 0
5 U 0
6 G 0
7 A 0
8 A 0
9 G 25
10 A 24
11 G 23
12 U 22
13 U 21
14 U 0

So, header of four lines (Filename, Organism, Accession Number, Citation)
Sequence, structure information in tuples of residue position, residue name,
residue partner. The residue partner is 0 if the base is unpaired.

Numbering is 1-based!
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from builtins import map
from builtins import str
from builtins import range
from .rna2d import Pairs


class BpseqParseError(Exception):
    """Exception raised when an error occurs during parsing a bpseq file"""
    pass


def parse_header(header_lines):
    """Return Info object from header information.

    header_lines -- list of lines or anything that behaves like it.

    Parses only the first three header lines with Filename, Organism, and
    Accession number. In general lines that contain a colon will be parsed.
    There's no error checking in here. If it fails to split on ':', the
    information is simply not added to the dictionary. The expected format
    for header lines is "key: value". The citation lane is parsed differently.
    """
    info = {}
    for line in header_lines:
        if line.startswith('Citation'):
            info['Citation'] = line.split()[-1].strip()
        elif ':' in line:
            try:
                field, value = map(str.strip, line.split(':', 1))
                info[field] = value
            except ValueError:
                # no interesting header line
                continue
        else:
            continue
    return info


def construct_sequence(seq_dict):
    """Construct RnaSequence from dict of {pos:residue}.

    seq_dict -- dictionary of {position: residue}

    Checks whether the first residue is 0. Checks whether all residues
    between min and max index are present. No checking on validity of
    residue symbols.
    """
    if not seq_dict:
        raise BpseqParseError("No sequence information")
    all_pos = seq_dict.keys()
    min_pos, max_pos = min(all_pos), max(all_pos)
    if min_pos != 0:
        raise BpseqParseError(
            "Something went wrong with adjusting the numbering")
    # make sure all positions are in the dictionary
    for idx in range(min_pos, max_pos + 1):
        if idx not in seq_dict:
            raise BpseqParseError(
                "Description of residue with index %s is missing" % (idx))
    seq = []
    for idx in range(min_pos, max_pos + 1):
        seq.append(seq_dict[idx])
    # Return as a simple string
    return ''.join(seq)


def parse_residues(residue_lines, num_base, unpaired_symbol):
    """Return RnaSequence and Pairs object from residue lines.

    residue_lines -- list of lines or anything that behaves like it.
        Lines should contain:
        residue_position, residue_identiy, residue_partner.
    num_base -- int, basis of the residue numbering. In bpseq files from
        the CRW website, the numbering starts at 1.
    unpaired_symbol -- string, symbol in the 'partner' column that indicates
        that a base is unpaired. In bpseq files from the CRW website, the
        unpaired_symbol is '0'. This parameter should be a string to allow
        other symbols that can't be casted to an integer to indicate
        unpaired bases.

    Checks for double entries both in the sequence and the structure, and
    checks that the structre is valid in the sense that if (up,down) in there,
    that (down,up) is the same.
    """
    # create dictionary/list for sequence and structure
    if not residue_lines:
        raise BpseqParseError("No residue lines in record")
    seq_dict = {}
    pairs = Pairs()

    for line in residue_lines:
        try:
            pos, res, partner = line.strip().split()
            if partner == unpaired_symbol:
                # adjust pos, not partner
                pos = int(pos) - num_base
                partner = None
            else:
                # adjust pos and partner
                pos = int(pos) - num_base
                partner = int(partner) - num_base
            pairs.append((pos, partner))

            # fill seq_dict
            if pos in seq_dict:
                raise BpseqParseError(
                    "Double entry for residue %s (%s in bpseq file)"
                    % (str(pos), str(pos + 1)))
            else:
                seq_dict[pos] = res

        except ValueError:
            raise BpseqParseError("Failed to parse line: %s" % (line))

    # check for conflicts, remove unpaired bases
    if pairs.hasConflicts():
        raise BpseqParseError("Conflicts in the list of basepairs")
    pairs = pairs.directed()
    pairs.sort()

    # construct sequence from seq_dict
    #seq = RnaSequence(construct_sequence(seq_dict))
    seq = construct_sequence(seq_dict)

    return seq, pairs


def MinimalBpseqParser(lines):
    """Separate header and content (residue lines).

    lines -- a list of lines or anything that behaves like that.

    The standard bpseq header (from the CRW website) is recognized. Also,
    lines that contain a colon are accepted as header lines.
    Header lines that aren't accepted as header, but that can be split into
    three parts are residue lines (sequence and structure description).
    Lines that don't fall into any of these categories are ignored.
    """
    result = {'HEADER': [], 'SEQ_STRUCT': []}

    for line in lines:
        if line.startswith('Filename') or line.startswith('Organism') or\
                line.startswith('Accession') or line.startswith('Citation') or\
                ":" in line:
            result['HEADER'].append(line.strip())
        elif len(line.split()) == 3:
            result['SEQ_STRUCT'].append(line.strip())
        else:
            continue  # unknown
    return result


def BpseqParser(lines, num_base=1, unpaired_symbol='0'):
    """Return RnaSequence and structure (Pairs object) specified in file.

    lines -- filestream of bpseq file. File should contain a single record.
    num_base -- int, basis of the residue numbering. In bpseq files from
        the CRW website, the numbering starts at 1.
    unpaired_symbol -- string, symbol in the 'partner' column that indicates
        that a base is unpaired. In bpseq files from the CRW website, the
        unpaired_symbol is '0'. This parameter should be a string to allow
        other symbols that can't be casted to an integer to indicate
        unpaired bases.

    Bpseq file looks like this:

    Filename: d.16.b.E.coli.bpseq
    Organism: Escherichia coli
    Accession Number: J01695
    Citation and related information available at http://www....
    1 A 0
    2 A 0
    3 A 0
    4 U 0
    5 U 0
    6 G 0
    7 A 0
    8 A 0
    9 G 25
    10 A 24
    11 G 23
    12 U 22
    13 U 21

    So, 4 header lines, followed by a list of residues.
    Position (indexed to 1), residue, partner position
    """
    # separate header and residue lines
    grouped_lines = MinimalBpseqParser(lines)
    # parse header and seq/struct separately
    #header_info = parse_header(grouped_lines['HEADER'])
    header_info = grouped_lines['HEADER']
    seq, struct = parse_residues(grouped_lines['SEQ_STRUCT'],
                                 num_base, unpaired_symbol)
    # add header info to the sequence as Info object
    #seq.Info = header_info

    return header_info, seq, struct

# ============================================================================
# CONVENIENCE FUNCTIONS
# ============================================================================


def bpseq_specify_output(lines, num_base=1, unpaired_symbol='0',
                         return_vienna=False, ):
    """Return Vienna structure of Pairs object with or without pseudoknots

    lines -- filestream of bpseq file. File should contain a single record.
    num_base -- int, basis of the residue numbering. In bpseq files from
        the CRW website, the numbering starts at 1.
    unpaired_symbol -- string, symbol in the 'partner' column that indicates
        that a base is unpaired. In bpseq files from the CRW website, the
        unpaired_symbol is '0'. This parameter should be a string to allow
        other symbols that can't be casted to an integer to indicate
        unpaired bases.
    return_vienna -- boolean, if True, a ViennaStructure object is returned,
        if False, a Pairs object is returned. If return_vienna is True,
        pseudoknots need to be removed from the structure.

    """
    _, seq, pairs = BpseqParser(lines, num_base, unpaired_symbol)

    if return_vienna:
        v = pairs.toVienna(len(seq))
        return seq, v
    return seq, pairs


if __name__ == "__main__":
    from sys import argv
    _, seq, struct = BpseqParser(open(argv[1]))
    print(seq)
    print(seq.Info)
    print(struct)



"""
This module has functions which take a Bulge-Graph as input
and return a modified copy of it.

This is implemented as a class which is accessible as
BulgeGraph.transformed. This way we can properly inherit in CoarseGrainRNA
"""
import copy
import logging

from .sequence import MissingResidue, Sequence


log = logging.getLogger(__name__)


class _GCDummy(object):
    """
    Can be used in place of GraphConstruction for BG-Initialization
    """

    def __init__(self, defines, edges):
        self.defines = defines
        self.edges = edges


class BGTransformer(object):
    def __init__(self, bg):
        self.bg = bg

    def _without_elements(self, elems):
        """
        Return a copy of the BulgeGraph without the elements in elem.
        Their residues will be converted to missing residues.

        :param elems: A list of element names, e.g. ["s1","s2"]
        """
        raise NotImplementedError("This is still work in progress.")
        # We use the PDB numbering in new_defines, so we do not have to adjust indices if we remove
        # residues in the middle.
        resid_defines = {}
        for k, v in self.bg.defines:
            resid_defines[k] = list(map(self.bg.seq.to_resid, v))
        new_edges = copy.deepcopy(self.bg.edges)
        to_missing = []
        for elem in elems:
            if elem[0] != "i":
                raise NotImplementedError("TODO")
            else:
                stem1, stem2 = new_edges[elem]
                elem_define_a = list(
                    map(self.bg.seq.to_resid, self.bg.define_a(elem)))
                # Remove the iloop
                to_missing.extend(
                    self.bg.define_residue_num_iterator(elem, seq_ids=True))
                new_edges[stem1].remove(elem)
                new_edges[stem2].remove(elem)
                del new_edges[elem]
                # Merge stem2 into stem1
                new_edges[stem1] |= new_edges[stem2]
                del new_edges[stem2]
                if resid_defines[stem1][1] in elem_define_a:
                    # stem1 -IL - stem2
                    resid_defines[stem1][1] = resid_defines[stem2][1]
                    resid_defines[stem1][2] = resid_defines[stem2][2]
                else:
                    # stem2 - il - stem1
                    resid_defines[stem1][0] = resid_defines[stem2][0]
                    resid_defines[stem1][3] = resid_defines[stem2][3]
                del resid_defines[stem2]
                del resid_defines[elem]

    def condensed(self):
        """
        Return a condensed copy of the BulgeGraph.

        In the condensed BulgeGraph only the first (most 5-prime) nucleotide
        or base-pair of each element is retained, and the other nts/ base-pairs are
        converted to missing residues.
        In basepairs the first basepair contains the most 5' and most 3'
        nucleotide of the stem.
        """
        log.debug("Condensing BG with break-points %s",
                  self.bg.backbone_breaks_after)
        log.info("Condensing Graph %s", self.bg.to_dotbracket_string())
        new_defines = {}
        new_seqids = []
        new_seq = ""
        new_missing = self.bg.seq.with_missing.export_missing()
        new_i = 1
        for elem in self.bg.iter_elements_along_backbone():
            if not self.bg.defines[elem]:
                new_defines[elem] = []
            else:
                if elem in new_defines:  # Backwards strand for stems:
                    if len(new_defines[elem]) != 2:
                        log.error("%s", self.bg.edges[elem])
                    assert len(new_defines[elem]) == 2, "{} doesn't have len 2".format(
                        new_defines[elem])
                    assert elem[0] in "si"
                    fr, to = self.bg.defines[elem][2:]
                    if elem[0] == "s":
                        # to keep basepairing consistent, we keep the last (not the first)
                        # nt.
                        keep_i = to
                    else:
                        # To keep interiour loops consistent with the case where
                        # it is only at the second strand, keep the first nt here
                        keep_i = fr
                    new_defines[elem].extend([new_i, new_i])
                    log.debug("Extended new_defines for %s to %s",
                              elem, new_defines[elem])
                else:
                    fr, to = self.bg.defines[elem][:2]
                    keep_i = fr
                    new_defines[elem] = [new_i, new_i]
                    log.debug("Set new_defines for %s to %s",
                              elem, new_defines[elem])
                new_i += 1
                new_seq += self.bg.seq[keep_i]
                new_seqids.append(self.bg.seq.to_resid(keep_i))
                for i in range(fr, to + 1):
                    if i != keep_i:
                        seq_id = self.bg.seq.to_resid(i)
                        new_missing.append(
                            MissingResidue(seq_id, self.bg.seq[i]))
                    if i in self.bg.backbone_breaks_after:
                        if i >= keep_i:
                            new_seq += "&"
                        else:
                            new_seq = new_seq[:-1] + "&" + new_seq[-1]
                        log.debug("Breakpoint %s: new_seq now %s", i, new_seq)
        log.info("Condensing iteration done. Now creating condensed BG")
        graph_constr = _GCDummy(new_defines, copy.deepcopy(self.bg.edges))
        seq = Sequence(new_seq, new_seqids, new_missing,
                       self.bg.seq._modifications)
        return type(self.bg)(graph_constr, seq, name=self.bg.name + "_condensed", _dont_split=True)



from collections import defaultdict
import logging
import itertools as it

from ..utilities.exceptions import GraphIntegrityError

log = logging.getLogger(__name__)

try:
    profile
except NameError:
    profile = lambda x: x


class BaseGraph(object):
    """
    A Base-class for the BulgeGraph and BulgeGraphConstruction.

    In contrast to BulgeGraph, this class does not enforce any nomenclature
    on the nodes.

    It does not support backbone breaks!

    It has no sequence.
    """

    def __init__(self):
        self.defines = {}
        self.edges = defaultdict(set)

    def connections(self, bulge):
        """
        :param g: Graph-like: A BulgeGraph or BulgeGraphConstruction.
        """
        def sort_key(x):
            if self.defines[x] and self.defines[x][0] == 1:
                # special case for stems at the beginning since there is no
                # adjacent nucleotide 0
                return 0
            return self.define_a(x)[0]
        connections = list(self.edges[bulge])
        connections.sort(key=sort_key)
        return connections

    def define_a(self, elem):
        """
        Returns the element define using the adjacent nucleotides (if present).

        If there is no adjacent nucleotide due to a chain break or chain end,
        then the nucleotide that is part of the element will be used instead.

        For stems, this always returns a list of length 4,
        for interior loops a list of length 2 or 4 and
        for other elements always a list of length 2.


        :param elem: An element name
        :returns: A list of integers
        """
        if self.defines[elem] == []:
            return self._define_a_zerolength(elem)
        return self._define_a_nonzero(elem)

    @profile
    def _define_a_nonzero(self, elem):
        define = self.defines[elem]
        new_def = []
        for i in range(0, len(define), 2):
            new_def.append(max(define[i] - 1, 1))
            new_def.append(define[i + 1] + 1)
        return new_def

    def flanking_nucleotides(self, d):
        '''
        Return the nucleotides directly flanking an element.

        :param d: the name of the element
        :return: a list of nucleotides
        '''
        set_adjacent = set(self.define_a(d))
        set_not_adjacent = set(self.defines[d])
        flanking = list(sorted(set_adjacent - set_not_adjacent))
        log.debug("Flanking nts are %s - %s = %s",
                  set_adjacent, set_not_adjacent, flanking)
        return flanking

    def _define_a_zerolength(self, elem):  # TODO speed-up by caching
        """
        Return the define with adjacent nucleotides for a zero-length element.

        Hereby we define that in cases of ambiuigity, the alphabetically first
        zero-length element comes at the lowest nucleotide position etc.

        :param elem: An element, e.g. "m0
        """
        if self.defines[elem] != []:
            raise ValueError("{} does not have zero length".format(elem))
        edges = self.edges[elem]
        if len(edges) == 1:  # Hairpin
            stem, = edges
            define = self.defines[stem]
            if define[2] == define[1] + 1:
                return [define[1], define[2]]
            raise GraphIntegrityError("Very strange zero-length hairpin {} "
                                      "(not?) connected to {}".format(elem, stem))
        if len(edges) == 2:
            stem1, stem2 = edges
            # See if this is the only element connecting the two stems.
            connections = self.edges[stem1] & self.edges[stem2]
            log.debug("Stems %s and %s, connected by %s have the following common edges: %s with defines %s",
                      stem1, stem2, elem, connections, list(map(lambda x: self.defines[x], connections)))
            zl_connections = []
            for conn in connections:
                if self.defines[conn] == []:
                    zl_connections.append(conn)
            assert elem in zl_connections
            # We DEFINE the 0-length connections to be sorted alphabetically by position
            zl_connections.sort()
            log.debug("Getting zl-coordinates")
            zl_coordinates = self._zerolen_defines_a_between(stem1, stem2)
            log.debug("Comparing zl-coordinates %s with connections %s",
                      zl_coordinates, zl_connections)
            if len(zl_connections) != len(zl_coordinates):
                raise GraphIntegrityError("Expecting stems {} and {} to have {} zero-length "
                                          "connections at nucleotide positions {}, however, "
                                          "found {} elements: {}".format(stem1, stem2,
                                                                         len(zl_coordinates),
                                                                         zl_coordinates,
                                                                         len(zl_connections),
                                                                         zl_connections))
            zl_coordinates = list(zl_coordinates)
            zl_coordinates.sort()
            i = zl_connections.index(elem)
            return list(zl_coordinates[i])
        raise GraphIntegrityError("Very strange zero length bulge {} with more than 2 adjacent "
                                  "elements: {}.".format(elem, edges))

    def _zerolen_defines_a_between(self, stem1, stem2):
        zl_coordinates = set()
        for k, l in it.product(range(4), repeat=2): # pylint: disable=W1638
            if abs(self.defines[stem1][k] - self.defines[stem2][l]) == 1:
                d = [self.defines[stem1][k], self.defines[stem2][l]]
                d.sort()
                log.debug("Zero-length element found: %s", d)
                zl_coordinates.add(tuple(d))
        log.debug("Returning zl-coordinates: %s", zl_coordinates)
        return zl_coordinates

    def _get_sides_plus(self, s1, bulge):
        """
        Get the side of s1 that is next to b.

        s1e -> s1b -> b

        :param s1: The stem.
        :param b: The bulge.
        :return: A tuple indicating the corner of the stem that connects
                 to the bulge as well as the corner of the bulge that connects
                 to the stem.
                 These sides are equivalent to the indices of the define.
        """
        if bulge not in self.edges[s1]:
            raise ValueError(
                "_get_sides_plus expects stem to be connected to bulge!")

        s1d = self.defines[s1]
        bd = self.defines[bulge]  # pylint: disable=C0103

        if not bd:
            bd = self._define_a_zerolength(bulge)  # pylint: disable=C0103
            bd[0] += 1
            bd[1] -= 1

        # before the stem on the 5' strand
        if s1d[0] - bd[1] == 1:
            return (0, 1)
        # after the stem on the 5' strand
        if bd[0] - s1d[1] == 1:
            return (1, 0)
        # before the stem on the 3' strand
        if s1d[2] - bd[1] == 1:
            return (2, 1)
        # after the stem on the 3' strand
        if bd[0] - s1d[3] == 1:
            return (3, 0)

        raise GraphIntegrityError(
            "Faulty bulge {}:{} connected to {}:{}".format(bulge, bd, s1, s1d))

    def get_node_from_residue_num(self, base_num):
        """
        Iterate over the defines and see which one encompasses this base.
        """
        seq_id = False
        for key in self.defines:
            for drange in self.define_range_iterator(key):
                if drange[0] <= base_num <= drange[1]:
                    return key
        raise LookupError(
            "Base number {} not found in the defines {}.".format(base_num, self.defines))

    def define_range_iterator(self, node, adjacent=False):
        """
        Return the ranges of the nucleotides in the define.

        In other words, if a define contains the following: [1,2,7,8]
        The ranges will be [1,2] and [7,8].

        :param adjacent: Use the nucleotides in the neighboring element which
                         connect to this element as the range starts and ends.
        :return: A list of two-element lists
        """
        if adjacent:
            define = self.define_a(node)
        else:
            define = self.defines[node]

        if define:
            yield [define[0], define[1]]
            if len(define) > 2:
                yield [define[2], define[3]]




"""
This module contains code for classifying a coarse-grained geometry as A-Minor
interaction.

.. warning:: This is intended for low-resolution data , as it does not take the
             orientation of individual bases into account. If you have all-atom
             data, dedicated tools like FR3D will be more accurate.

If you just want to classify interactions in a given structure, you only need
the functions `classify_interaction` or `all_interactions`.

To train your own version of the classifier, modify its parameters or perform
cross-validation, use the AMinorClassifier.

Access the default trainings data with the get_trainings_data(loop_type)
function.
"""

from __future__ import print_function, division, absolute_import

import pkgutil
import logging
import warnings
import json

try:
    from io import StringIO
except ImportError:
    from StringIO import StringIO

import numpy as np

import pandas as pd

from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
try:
    from sklearn.neighbors import KernelDensity
except:
    from sklearn.neighbors.kde import KernelDensity

from sklearn.metrics import confusion_matrix

import forgi.threedee.utilities.vector as ftuv
import forgi.threedee.utilities.graph_pdb as ftug


log = logging.getLogger(__name__)


CUTOFFDIST = 30
ANGLEWEIGHT = 10



P_INTERACTION = 0.05  # 0.03644949066213922

################# Just classify my structure ############################


def loop_potential_interactions(cg, loop, domain=None):
    """
    Iterate over all stems and return those loop-stem pairs that will be passed
    to the AMinor classification and are not ruled out beforehand.
    """
    geos = []
    labels = []
    if domain is not None:
        stems = (s for s in domain if s[0] == "s")
    else:
        stems = cg.stem_iterator()
    for stem in stems:
        if stem in cg.edges[loop]:
            continue
        # To save computation time
        if not ftuv.elements_closer_than(cg.coords[loop][0],
                                         cg.coords[loop][1],
                                         cg.coords[stem][0],
                                         cg.coords[stem][1],
                                         CUTOFFDIST):
            continue
        geos.append(get_relative_orientation(cg, loop, stem))
        labels.append([loop, stem])
    geos = np.array(geos)
    if len(geos) > 0:
        geos[:, 0] /= ANGLEWEIGHT
    return geos, labels


def potential_interactions(cg, loop_type, domain=None):
    """
    :returns: A tuple `geos`, `labels`.
              `geos` is an Nx3 array, where N is the number of
              potential interactions and the inner dimension is dist, angle1, angle2.
              `geos` can be passed to the AMinor classifier.
              `labels` is an Nx2 array, where the inner dimension is loopname, stemname
    """
    labels = []
    geos = []
    for loop in cg.defines:
        if domain is not None and loop not in domain:
            continue
        if loop[0] != loop_type:
            continue
        if 'A' not in "".join(cg.get_define_seq_str(loop)):
            continue
        loop_geos, loop_labels = loop_potential_interactions(cg, loop, domain)
        geos.extend(loop_geos)
        labels.extend(loop_labels)
    return np.array(geos), np.array(labels)


def all_interactions(cg, clfs=None):
    """
    Get a list of all predicted A-Minor interactions in a cg-object.

    This is more efficient than using classify_interaction iteratively,
    because it uses vectorization.

    :param clfs: A dictionary {loop_type: AMinorClassifier} where
                 loop_type is one of "i", "h", "m".
                 If clfs is None or a key is missing, uses the default
                 pretrained classifier.

    :returns: A list of  tuples (loop, stem)
    """
    interactions = []
    for loop_type in ["i", "h"]:
        if clfs is not None and loop_type not in clfs:
            warnings.warn("No classifier specified for loop type %s (only for %s), "
                          "using default classifier.", loop_type, ",".join(clfs.keys()))
        if clfs is None or loop_type not in clfs:
            clf = _get_default_clf(loop_type)
        else:
            clf = clfs[loop_type]
        geos, labels = potential_interactions(cg, loop_type)
        interactions.extend(
            _classify_potential_interactions(clf, geos, labels))
    return interactions


def classify_interaction(cg, loop, stem=None, clf=None):
    """
    Returns the interaction pair loop, stem as a tuple or False if no interaction exists.

    :param cg: The CoarseGrainRNA
    :param loop: The loop name, e.g. "i0"
    :param stem: A stem name, e.g. "s0" to consider interactions of loop with another stem as False
                 and only an interaction with this stem as True.

                ..warning:: Our statistical modelling allows for at most 1 interaction per loop.
                            This means that we have calculate the interaction probability of this
                            loop with all stems, even if stem is given. If another stem has a
                            higher interaction probability than the given stem,
                            this function will return False, regardless of the interaction
                            probability stem-loop.

    :param clf: A trained AMinorClassifier or None (use default classifier for loop type)


    ..note:: `all_interactions` is more efficient, if you are interested in all loop-stem pairs.

    :returns: A tuple (loop, stem) or False
    """
    if clf is None:
        clf = _get_default_clf(loop[0])
    geos, labels = loop_potential_interactions(cg, loop)
    interactions = _classify_potential_interactions(clf, geos, labels)
    if not interactions:
        return False
    if stem is not None and interactions != [(loop, stem)]:
        return False
    interaction, = interactions
    return interaction

############# Roll your own classifier ###########################


def get_trainings_data(loop):
    loop_type = loop[0]
    return _DefaultClf._get_data(loop_type)


class AMinorClassifier(BaseEstimator, ClassifierMixin):
    """
    A classifier that predicts A-Minor interactions based on the following formula:

    .. math:: P(I|geo)=f(geo|I)/f(geo)*P(I)

    where :math:`f` is a probability density and :math:`P` is a probability
    and :math:`I` means interaction.

    Since there are no interactions with more than 30 Angstrom distance,
    interactions require an A in the loop sequence and estimating numerator
    and denominator seperately could lead to probabilities greater than 1,
    we use the following formula:

    .. math:: P(I| (geo, d<30, A in seq)) = num/denom

    where
    .. math::

        num=f(geo|I)*P(I|(d<30, A \\in seq))

        denom=num+f(geo|(\\not I, d<30))*(1-P(I|(d<30, A \\in seq))

    We estimate the probability densities as kernel density estimates.
    For :math:`f(geo|I)` we use annotations created by FR3D-searches for
    all 4 types of A-minor motifs.
    We assume that :math:`f(geo|(\\not I, A \\in seq))== f(geo|(\\not I, A \\in seq))`
    and that FR3D might miss some true interactions. Thus we estimate
    :math:`f(geo|(\\not I, A \\in seq, d<30))` as
    :math:`f(geo|(\\not I, A \\notin seq, d<30))`

    Since both densities are normalized, it does not matter that we use a
    different number of datapoints for their estimation.

    We estimate P(I|(d<30, A \\in seq)) as the number of occurrences where
    a loop with A is in an A-minor interaction, over the number of all
    loop-stem pairs with less than 30 angstrom distance and an A in the sequence.
    """

    def __init__(self, kernel="linear", bandwidth=0.3, symmetric=True, p_I=P_INTERACTION):
        self.p_I = p_I
        self.symmetric = symmetric
        self.kernel = kernel
        self.bandwidth = bandwidth

    def fit(self, X, y):
        """
        Train the model.

        :param X: A Nx3 array, where the features are
                  distance(Angstrom)/10, angle1(rad), angle2(rad)
                  The **distance** is the closest distance between the two line
                  segments (i.e. coarse grained elementts)
                  **angle1** is the angle between the line along the stem vector
                  and the line along the shortest connection between the two
                  elements. A an angle between two straight lines, it is
                  defined between 0 and 90 degrees.
                  **angle2** is the angle between the connecting vector
                  (pointing from the stem to the loop), projected onto the
                  plane normal to the stem direction and the twist vector
                  (location of minor groove) at the point closest to the
                  interaction. As an angle between two vectors, it is
                  defined between 0 and 180 degrees.
        :param y: An array of length N. 0 means no interaction,
                  1 means interaction.
        """
        # Check that X and y have correct shape
        X, y = check_X_y(X, y)
        log.info("Trainings-data has shape %s", X.shape)
        log.info("We have %s known interactions ", sum(y))
        if X.shape[1] != 3:
            raise TypeError(
                "Expect exactly 3 features, found {}".format(X.shape[1]))
        if not all(yi in [0, 1] for yi in y):
            raise ValueError("y should only contain the values 1 and 0")
        ame = X[np.where(y)]
        non_ame = X[np.where(y == 0)]
        if self.symmetric:
            ame = self._make_symmetric(ame)
            non_ame = self._make_symmetric(non_ame)
        log.info("Fitting. First positive sample: %s", X[np.where(y)][0])
        self.ame_kde_ = KernelDensity(kernel=self.kernel,
                                      bandwidth=self.bandwidth).fit(ame).score_samples
        self.non_ame_kde_ = KernelDensity(kernel=self.kernel,
                                          bandwidth=self.bandwidth).fit(non_ame).score_samples
        self.X_ = X
        self.y_ = y

    @staticmethod
    def _make_symmetric(geos):
        """
        Make the trainingsdata symmetric around those bounds of the radial
        variables, where we expect a high density.

        Kernel density estimation does not work too well out of the box on
        bounded support: Samples close to the boundary have create a
        significant density outside the support. Due to symmetry considerations
        of angular data, we can avoid this issue by adding symmetric datapoints
        the following way: Mirror all datapoints along the plane
        angle1=180 degrees (both angles between the two lines are equivalent) and
        around the plane angle2==0 degrees (both rotational
        directions are equivalent).
        By creating 4 times as many datapoints, we avoid bias near the boundary.
        """
        geos = np.concatenate([geos,
                               [(d, np.pi - a1, a2) for d, a1, a2 in geos]])
        # We allow negative angles for angle2, so there is no need to make it symmetric.
        # geos = np.concatenate([geos,
        #                       [(d, a1, -a2) for d, a1, a2 in geos]])
        return geos

    def predict(self, X):
        return self.predict_proba(X) > 0.5

    def score(self, X, y):
        """
        The average between specificity and sensitivity
        """
        y_pred = self.predict(X)
        tn, fp, fn, tp = confusion_matrix(y, y_pred).ravel()
        specificity = tn / (tn + fp)
        sensitivity = tp / (tp + fn)
        return (0.8 * sensitivity + 0.2 * specificity)

    def predict_proba(self, X):
        check_is_fitted(self, ['ame_kde_', 'non_ame_kde_'])
        X = check_array(X)
        log.debug("Predicting for %s", X)
        numerator = np.exp(self.ame_kde_(X)) * self.p_I
        denom = numerator + np.exp(self.non_ame_kde_(X)) * (1 - self.p_I)
        with warnings.catch_warnings():  # division by 0
            warnings.simplefilter("ignore", RuntimeWarning)
            return np.nan_to_num(numerator / denom)

    def set_params(self, **kwargs):
        """"""
        super(AMinorClassifier, self).set_params(**kwargs)
        # If it was fitted, we must propagate the parameter changes
        # to the child-KDEs by refitting to the same data
        if hasattr(self, "X_"):
            self.fit(self.X_, self.y_)
        return self


############## get orientation #########################

def get_loop_flexibility(cg, loop):
    """
    Unused. We tried to see if the length of the loop vs # bases had an effect on ointeraction probability.
    """
    assert loop[0] == "i"
    d = cg.define_a(loop)
    nt1, nt2 = d[1] - d[0], d[3] - d[2]
    max_nts = max(nt1, nt2)
    loop_length = ftuv.magnitude(cg.coords.get_direction(loop))
    # As number of nucleotide-links (or phosphate groups) per Angstrom
    # 9.2 is the sum of average bond lengths for bonds in the nucleotide linkage.
    # Bond lengths taken from: DOI: 10.1021/ja9528846
    # A value of 1 means, all bonds are stretched.
    # Ideal helices have a value of: 4.41
    # A value below 1 should be rare.
    # Higher values mean higher flexibility.
    return (max_nts) / loop_length * 9.2


def get_relative_orientation(cg, loop, stem):
    '''
    Return how loop is related to stem in terms of three parameters.

    The stem is the receptor of a potential A-Minor interaction, whereas the
    loop is the donor.

    The 3 parameters are:

        1.  Distance between the closest points of the two elements
        2.  The angle between the stem and the vector between the two
        3.  The angle between the minor groove of l2 and the projection of
            the vector between stem and loop onto the plane normal to the stem
            direction.
    '''
    point_on_stem, point_on_loop = ftuv.line_segment_distance(cg.coords[stem][0],
                                                              cg.coords[stem][1],
                                                              cg.coords[loop][0],
                                                              cg.coords[loop][1])
    conn_vec = point_on_loop - point_on_stem
    dist = ftuv.magnitude(conn_vec)
    angle1 = ftuv.vec_angle(cg.coords.get_direction(stem),
                            conn_vec)
    # The direction of the stem vector is irrelevant, so
    # choose the smaller of the two angles between two lines
    if angle1 > np.pi / 2:
        angle1 = np.pi - angle1
    tw = cg.get_twists(stem)
    if dist == 0:
        angle2 = float("nan")
    else:
        if stem[0] != 's':
            raise ValueError(
                "The receptor needs to be a stem, not {}".format(stem))
        else:
            stem_len = cg.stem_length(stem)
            # Where along the helix our A-residue points to the minor groove.
            # This can be between residues. We express it as floating point nucleotide coordinates.
            # So 0.0 means at the first basepair, while 1.5 means between the second and the third basepair.
            pos = ftuv.magnitude(point_on_stem - cg.coords[stem][0]) / ftuv.magnitude(
                cg.coords.get_direction(stem)) * (stem_len - 1)
            # The vector pointing to the minor groove, even if we are not at a virtual residue (pos is a float value)
            virt_twist = ftug.virtual_res_3d_pos_core(
                cg.coords[stem], cg.twists[stem], pos, stem_len)[1]
            # The projection of the connection vector onto the plane normal to the stem
            conn_proj = ftuv.vector_rejection(
                conn_vec, cg.coords.get_direction(stem))
            try:
                # Note: here the directions of both vectors are well defined,
                # so angles >90 degrees make sense.
                angle2 = ftuv.vec_angle(virt_twist, conn_proj)
            except ValueError:
                if np.all(virt_twist == 0):
                    angle2 = float("nan")
                else:
                    raise
            # Furthermore, the direction of the second angle is meaningful.
            # We call use a positive angle, if the cross-product of the two vectors
            # has the same sign as the stem vector and a negative angle otherwise
            cr = np.cross(virt_twist, conn_proj)
            sign = ftuv.is_almost_parallel(cr,  cg.coords.get_direction(stem))
            #assert sign != 0, "{} vs {} not (anti) parallel".format(
            #    cr, cg.coords.get_direction(stem))
            angle2 *= sign

    return dist, angle1, angle2

############# Private ##########################################


def _get_masks(df, loop_type):
    lt_mask = df.loop_type == loop_type
    cutoff_mask = df.dist < CUTOFFDIST
    in_support = lt_mask & cutoff_mask
    has_a = df.loop_sequence.str.contains("A")
    is_i = df.is_interaction
    mask_ame = in_support & is_i & has_a
    mask_non_ame = in_support & np.invert(is_i | has_a)
    mask_non_fred = in_support & np.invert(is_i) & has_a
    return mask_ame, mask_non_ame, mask_non_fred


def _classify_potential_interactions(clf, geos, labels):
    if len(geos) == 0:
        return []
    score = clf.predict_proba(geos)
    y = score > 0.5
    log.info("Classifying %s", labels)
    log.info("score %s", score)
    log.info("# hits (not unique)=%s", sum(y))
    # We modelled the probabilities in a way that each loop can have only 1 interaction.
    # So for multiple interactions, use only the best one.
    best_interactions = {}
    for i, label in enumerate(labels):
        loop = label[0]
        if score[i] < 0.5:
            continue
        if loop not in best_interactions or score[i] > best_interactions[loop][0]:
            best_interactions[loop] = (score[i], label)
    log.info("%s unique interacting loops.", len(best_interactions))
    log.debug(best_interactions)
    return [best_i[1] for best_i in best_interactions.values()]


class _DefaultClf(object):
    """Just to put global (cached) variables into a seperate scope."""
    _clfs = {}  # The pretrained classifiers, lazily loaded.
    _geo_df = None

    @classmethod
    def get_default_clf(cls, loop_type):
        if loop_type not in cls._clfs:
            clf = AMinorClassifier()
            rawdata = pkgutil.get_data(
                'forgi', 'threedee/data/aminor_params.json')
            params = json.load(StringIO(rawdata.decode("ascii")))
            clf.set_params(**params[loop_type])
            X, y = cls._get_data(loop_type)
            clf.fit(X, y)
            cls._clfs[loop_type] = clf
        return cls._clfs[loop_type]

    @classmethod
    def get_dataframe(cls):
        if cls._geo_df is None:
            rawdata = pkgutil.get_data(
                'forgi', 'threedee/data/aminor_geometries.csv')
            cls._geo_df = pd.read_csv(
                StringIO(rawdata.decode("ascii")), comment="#", sep=" ")
        return cls._geo_df

    @classmethod
    def _get_data(cls, loop_type):
        df = cls.get_dataframe()
        return df_to_data_labels(df, loop_type)


def _get_default_clf(loop):
    """
    :param loop: A element name, e.g. "i0" or a loop-type, e.g. "i"
    """
    loop_type = loop[0]
    return _DefaultClf.get_default_clf(loop_type)


def df_to_data_labels(df, loop_type):
    """
    Create the trainings data as two arrays X and y (or data and labels)
    from the initial dataframe

    :returns: X, y
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        df["loop_name"] = df.pdb_id + \
            df.interaction.str.split("-").apply(lambda x: x[0])
    mask_ame, mask_non_ame, mask_non_fred = _get_masks(df, loop_type)
    data = pd.concat([df[mask_ame].drop_duplicates(["loop_name"]),
                      df[mask_non_ame]])
    data = np.array([[x.dist / ANGLEWEIGHT, x.angle1, x.angle2, x.is_interaction]
                     for x in data.itertuples()])
    return data[:, :3], data[:, 3]



import sys
from collections import namedtuple, defaultdict
"""
Code by me (B.Thiel), borrowed from another, unreleased project.
"""


class LineBasedParser(object):
    comment = "#"
    skip_empty = True
    strip = True
    """
    A baseclass for line-based parsers.

    It provides the member functions `parse` taking a filename,
    and `parse_string`, taking a multiline string.

    If you subclass this class, you have to implement `_parse_line`
    and can implement `_before_parsing` and `_after_parsing`

    The class-level variables `comment`, `skip_empty` and `strip`
    controll how lines without content are handled.

    * If strip is True, all lines are stripped before further processing.
    * If skip_empty is True, empty lines (after stripping) are skipped
      and not passed to _parse_line
    * If comment is non-empty, all lines starting with the comment character(s)
      (after stripping) are skipped as well.
    """

    def parse(self, filename):
        """
        :param filename: A filename. This file will be opened for reading.
        :returns: self.result (has to be filled by self._parse_line)
        """
        with open(filename) as f:
            return self._parse(f)

    def parse_string(self, string):
        """
        :param string: A multiline string
        :returns: see LineBasedParser.parse
        """
        return self._parse(string.splitlines())

    def _before_parsing(self):
        pass

    def _parse_line(self, line):
        pass

    def _after_parsing(self):
        pass


    def _parse(self, lines):
        """
        This function should not be overridden.

        This function handles reading of lines, comments, empty-lines.
        """
        self.result = None
        self._before_parsing()
        for i, line in enumerate(lines):
            if self.strip:
                line = line.strip()
            if self.skip_empty and not line:
                continue
            if self.comment and line.startswith(self.comment):
                continue
            self._parse_line(line)
        self._after_parsing()
        return self.result


class ChainIdMappingParser(LineBasedParser):
    comment = ""  # No comment characters allowed in these files.

    def _before_parsing(self):
        self.result = namedtuple("ChainChainMapping", [
                                 "bundle2mmcif", "mmcif2bundle"])(defaultdict(dict), {})
        self.bundle = None

    def _parse_line(self, line):
        if self.bundle is None:
            if "New chain ID" not in line or "Original chain ID" not in line:
                raise ValueError("Expecting first line to contain "
                                 "'New chain ID' and 'Original chain ID'")
            else:
                self.bundle = "HEADER"
                return
        elif self.bundle == "HEADER" or "-pdb-bundle" in line:
            if not line[4:].startswith("-pdb-bundle"):
                raise ValueError("Expecting '-pdb-bundle'", charno=4)
            else:
                if not line[-1] == ":":
                    raise ValueError(
                        "Expecting pdb-bundle name to be followed by ':'")
                self.bundle = line[:-1]
                return
        else:
            new, old = line.split()
            new = new.strip()
            old = old.strip()
            if len(new) != 1:
                raise ValueError(
                    "Expecting chain in pbd-bundle to be only 1 character.")
            self.result.bundle2mmcif[self.bundle][new] = old
            self.result.mmcif2bundle[old] = (self.bundle, new)
            return

    def _after_parsing(self):
        # Convert the defaultdict to a normal dict.
        self.result = namedtuple("ChainChainMapping",
                                 ["bundle2mmcif", "mmcif2bundle"])(dict(self.result.bundle2mmcif), self.result.mmcif2bundle)
